from collections import Counter
from collections.abc import Sequence
from functools import partial
from pathlib import Path

from rich.console import Console
from rich.markup import escape

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.http_client import (
    HTTPClient,
    ItemsFailedRequest,
    ItemsFailedResponse,
    ItemsResultMessage,
    ItemsSuccessResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling import ViewReference
from cognite_toolkit._cdf_tk.constants import DATA_MANIFEST_SUFFIX, DATA_RESOURCE_DIR
from cognite_toolkit._cdf_tk.cruds import ViewCRUD
from cognite_toolkit._cdf_tk.exceptions import ToolkitRuntimeError, ToolkitValueError
from cognite_toolkit._cdf_tk.protocols import T_ResourceRequest, T_ResourceResponse
from cognite_toolkit._cdf_tk.storageio import (
    FileContentIO,
    T_Selector,
    UploadableStorageIO,
    get_upload_io,
)
from cognite_toolkit._cdf_tk.storageio._base import TableUploadableStorageIO, UploadItem
from cognite_toolkit._cdf_tk.storageio.selectors import Selector, load_selector
from cognite_toolkit._cdf_tk.storageio.selectors._instances import InstanceSpaceSelector, InstanceViewSelector
from cognite_toolkit._cdf_tk.tk_warnings import HighSeverityWarning, MediumSeverityWarning, ToolkitWarning
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables
from cognite_toolkit._cdf_tk.utils.fileio import MultiFileReader
from cognite_toolkit._cdf_tk.utils.producer_worker import ProducerWorkerExecutor
from cognite_toolkit._cdf_tk.utils.progress_tracker import ProgressTracker
from cognite_toolkit._cdf_tk.utils.useful_types import JsonVal

from ._base import ToolkitCommand
from .deploy import DeployCommand


class UploadCommand(ToolkitCommand):
    """Command for uploading data to CDF from files in a specified directory.

    This command reads files matching a specific pattern in the input directory,
    processes them using a StorageIO instance, and uploads the data to CDF.

    Attributes:
        _MAX_QUEUE_SIZE (int): The maximum size of the queue for processing items.
            Set to 80 to balance memory usage and processing speed.
    """

    _MAX_QUEUE_SIZE = 80
    _MAX_VERBOSE_PRINTED_FAILED_IDS = 10
    _UPLOAD = "upload"

    def upload(
        self,
        input_dir: Path,
        client: ToolkitClient,
        deploy_resources: bool,
        dry_run: bool,
        verbose: bool,
        kind: str | None = None,
    ) -> None:
        """Uploads data from files in the specified input directory to CDF.

        Args:
            input_dir: The directory containing the files to upload. It is expected to
                have the structure defined below.
            client: An instance of ToolkitClient to interact with CDF.
            deploy_resources: If True, deploys resources from the 'resources' subdirectory.
            dry_run: If True, performs a dry run without actually uploading the data
                (or deploying resources).
            verbose: If True, prints detailed information about the upload process.
            kind: Optional; if provided, only data files of this kind will be processed.

        The expected structure of the input directory is as follows:
        ```
        input_dir/
        ├── resources/                # Optional, only if deploy_resources is True
        │   ├── raw/
        │   │   ├── table1.Table.yaml
        │   │   └── table2.Table.yaml
        │   └── ...
        ├── datafile1.kind.ndjson # Data file of a specific kind
        ├── datafile1.Manifest.yaml       # Manifest for datafile1
        ├── datafile2.kind2.ndjson # Another data file of the same or different kind
        ├── datafile2.Manifest.yaml       # Manifest file for datafile2
        └── ...
        """
        console = client.console
        data_files_by_selector = self._find_data_files(input_dir)

        self._deploy_resource_folder(input_dir / DATA_RESOURCE_DIR, deploy_resources, client, console, dry_run, verbose)

        data_files_by_selector = self._topological_sort_if_instance_selector(data_files_by_selector, client)
        self._upload_data(data_files_by_selector, client, dry_run, input_dir, console, verbose)

    def _topological_sort_if_instance_selector(
        self, data_files_by_selector: dict[Selector, list[Path]], client: ToolkitClient
    ) -> dict[Selector, list[Path]]:
        """Topologically sorts InstanceSelectors (if they are present) to determine the order of upload based on container dependencies from the views.

        Args:
            data_files_by_selector: A dictionary mapping selectors to their data files.
            client: The cognite client to use for the upload.

        Returns:
            A dictionary mapping selectors to their data files with necessary preprocessing.
        """
        counts = Counter(type(selector) for selector in data_files_by_selector.keys())
        if (counts[InstanceSpaceSelector] + counts[InstanceViewSelector]) <= 1:
            return data_files_by_selector

        selector_by_view_id: dict[ViewReference, Selector] = {}
        for selector in data_files_by_selector:
            if isinstance(selector, InstanceSpaceSelector | InstanceViewSelector) and selector.view is not None:
                view_ref = selector.view.as_id()
                if not isinstance(view_ref, ViewReference):
                    raise RuntimeError(
                        f"View {view_ref} does not have a version, which is required for topological sorting."
                    )
                selector_by_view_id[view_ref] = selector

        view_dependencies, cyclic_views = ViewCRUD.create_loader(client).topological_sort_container_constraints(
            list(selector_by_view_id.keys())
        )
        if cyclic_views:
            MediumSeverityWarning(
                f"Cyclic dependencies detected between views: {', '.join(str(view) for view in cyclic_views)}. "
                f"If their constraints are not already satisfied by instances in existing containers / views "
                f"in CDF, the upload may fail for these views."
            ).print_warning(console=client.console)
            view_dependencies = view_dependencies + cyclic_views

        prepared_selectors: dict[Selector, list[Path]] = {}

        # Reorder selectors according to the dependency-sorted view list
        for view_id in view_dependencies:
            selector = selector_by_view_id[view_id]
            prepared_selectors[selector] = data_files_by_selector[selector]

        # Preserve selectors that aren't affected by view dependencies
        # (e.g., raw tables, time series, non-view instance data)
        for selector in data_files_by_selector.keys():
            if selector not in prepared_selectors:
                prepared_selectors[selector] = data_files_by_selector[selector]

        return prepared_selectors

    def _find_data_files(
        self,
        input_dir: Path,
    ) -> dict[Selector, list[Path]]:
        """Finds data files and their corresponding metadata files in the input directory."""
        data_files_by_metadata: dict[Selector, list[Path]] = {}
        for manifest_file in input_dir.glob(f"*{DATA_MANIFEST_SUFFIX}"):
            selector_or_warning = load_selector(manifest_file)
            if isinstance(selector_or_warning, ToolkitWarning):
                self.warn(selector_or_warning)
                continue
            selector: Selector = selector_or_warning
            data_files = selector.find_data_files(input_dir, manifest_file)
            if not data_files:
                self.warn(
                    MediumSeverityWarning(
                        f"Metadata file {manifest_file.as_posix()!r} has no corresponding data files, skipping.",
                    )
                )
                continue
            data_files_by_metadata[selector] = data_files
        return data_files_by_metadata

    def _deploy_resource_folder(
        self,
        resource_dir: Path,
        deploy_resources: bool,
        client: ToolkitClient,
        console: Console,
        dry_run: bool,
        verbose: bool,
    ) -> None:
        """Deploy resources from the specified resource directory if it exists and deployment is enabled."""
        if deploy_resources and resource_dir.exists():
            deploy_command = DeployCommand()
            deploy_results = deploy_command.deploy_all_resources(
                client=client,
                build_dir=resource_dir,
                env_vars=EnvironmentVariables.create_from_environment(),
                dry_run=dry_run,
                verbose=verbose,
            )
            self.warning_list.extend(deploy_command.warning_list)
            if deploy_results.has_counts:
                console.print(deploy_results.counts_table())
        elif deploy_resources:
            self.warn(
                MediumSeverityWarning(
                    f"Resource directory {resource_dir.as_posix()!r} does not exist, skipping resource deployment."
                )
            )

    def _upload_data(
        self,
        data_files_by_selector: dict[Selector, list[Path]],
        client: ToolkitClient,
        dry_run: bool,
        input_dir: Path,
        console: Console,
        verbose: bool,
    ) -> None:
        total_file_count = sum(len(files) for files in data_files_by_selector.values())
        if verbose:
            input_dir_display = self._path_as_display_name(input_dir)
            console.print(f"Found {total_file_count} files to upload in {input_dir_display.as_posix()!r}.")
        action = "Would upload" if dry_run else "Uploading"
        with HTTPClient(config=client.config) as upload_client:
            file_count = 1
            for selector, datafiles in data_files_by_selector.items():
                io = self._create_selected_io(selector, datafiles[0], client)
                if io is None:
                    continue
                reader = MultiFileReader(datafiles)
                # FileContentIO supports uploading any file format.
                if reader.is_table and not isinstance(io, TableUploadableStorageIO | FileContentIO):
                    raise ToolkitValueError(f"{selector.display_name} does not support {reader.format!r} files.")

                item_count = io.count_items(reader, selector)
                iteration_count = item_count // io.CHUNK_SIZE + (1 if item_count % io.CHUNK_SIZE > 0 else 0)

                tracker = ProgressTracker[str]([self._UPLOAD])
                executor = ProducerWorkerExecutor[list[tuple[str, dict[str, JsonVal]]], Sequence[UploadItem]](
                    download_iterable=io.read_chunks(reader, selector),
                    process=partial(io.rows_to_data, selector=selector)
                    if reader.is_table and isinstance(io, TableUploadableStorageIO)
                    else io.json_chunk_to_data,
                    write=partial(
                        self._upload_items,
                        upload_client=upload_client,
                        io=io,
                        dry_run=dry_run,
                        selector=selector,
                        tracker=tracker,
                        console=console,
                        verbose=verbose,
                    ),
                    iteration_count=iteration_count,
                    max_queue_size=self._MAX_QUEUE_SIZE,
                    download_description=f"Reading {selector.display_name!r} files",
                    process_description="Processing",
                    write_description=f"{action} {selector.display_name!r}",
                    console=console,
                )
                executor.run()
                file_count += len(datafiles)
                executor.raise_on_error()
                final_action = "Uploaded" if not dry_run else "Would upload"
                suffix = " successfully" if not dry_run else ""
                results = tracker.aggregate()
                success = results.get((self._UPLOAD, "success"), 0)
                failed = results.get((self._UPLOAD, "failed"), 0)
                if failed > 0:
                    suffix += f", {failed:,} failed"
                console.print(
                    f"{final_action} {success:,} {selector.display_name} from {len(datafiles)} files{suffix}."
                )

    @staticmethod
    def _path_as_display_name(input_path: Path, cwd: Path = Path.cwd()) -> Path:
        display_name = input_path
        if input_path.is_relative_to(cwd):
            display_name = input_path.relative_to(cwd)
        return display_name

    def _create_selected_io(
        self, selector: Selector, data_file: Path, client: ToolkitClient
    ) -> UploadableStorageIO | None:
        try:
            io_cls = get_upload_io(selector)
        except ValueError as e:
            self.warn(HighSeverityWarning(f"Could not find StorageIO for selector {selector}: {e}"))
            return None
        return io_cls(client)

    @classmethod
    def _upload_items(
        cls,
        data_chunk: Sequence[UploadItem[T_ResourceRequest]],
        upload_client: HTTPClient,
        io: UploadableStorageIO[T_Selector, T_ResourceResponse, T_ResourceRequest],
        selector: T_Selector,
        dry_run: bool,
        tracker: ProgressTracker[str],
        console: Console,
        verbose: bool,
    ) -> None:
        if dry_run:
            for item in data_chunk:
                tracker.set_progress(item.source_id, cls._UPLOAD, "success")
            return
        results = io.upload_items(data_chunk, upload_client, selector)
        all_failed = True
        # Group failed ids by error description for cleaner output
        failures_by_error: dict[str, list[str]] = {}
        for message in results:
            if isinstance(message, ItemsSuccessResponse):
                all_failed = False
                for id_ in message.ids:
                    tracker.set_progress(id_, step=cls._UPLOAD, status="success")
            elif isinstance(message, ItemsResultMessage):
                for id_ in message.ids:
                    tracker.set_progress(id_, step=cls._UPLOAD, status="failed")
                if verbose:
                    if isinstance(message, ItemsFailedResponse):
                        error_description = f"(HTTP {message.status_code}): {escape(message.error.message)}"
                        failures_by_error.setdefault(error_description, []).extend(message.ids)
                    elif isinstance(message, ItemsFailedRequest):
                        failures_by_error.setdefault(escape(message.error_message), []).extend(message.ids)
            else:
                console.log(f"[red]Unexpected result from upload: {escape(str(message))}[/red]")
        if verbose:
            for error_description, failed_ids in failures_by_error.items():
                ids_display = escape(", ".join(failed_ids[: cls._MAX_VERBOSE_PRINTED_FAILED_IDS]))
                if len(failed_ids) > cls._MAX_VERBOSE_PRINTED_FAILED_IDS:
                    ids_display += f" ... and {len(failed_ids) - cls._MAX_VERBOSE_PRINTED_FAILED_IDS} more"
                console.print(
                    f"[red]Failed to upload[/red] {len(failed_ids)} items from [bold]{selector}[/bold] "
                    f"{error_description}\n"
                    f"  Failed items: {ids_display}"
                )
        if all_failed and results:
            raise ToolkitRuntimeError(
                "Upload process was stopped due to repeatedly failed uploads. Rerun with --verbose to see detailed failure information."
            )



from abc import ABC, abstractmethod
from collections.abc import Iterable, Sequence
from itertools import zip_longest
from typing import Any, Generic, Literal, overload

from pydantic import JsonValue

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, PagedResponse, ResponseItems
from cognite_toolkit._cdf_tk.client.cdf_client.api import APIMethod, Endpoint
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, RequestMessage, SuccessResponse
from cognite_toolkit._cdf_tk.client.request_classes.filters import InstanceFilter
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling import (
    InstanceRequest,
    InstanceResponse,
    QueryRequest,
)
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling._instance import InstanceSlimDefinition
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling._query import (
    QueryResponseTyped,
    QueryResponseUntyped,
)
from cognite_toolkit._cdf_tk.client.resource_classes.instance_api import (
    T_InstancesListRequest,
    T_InstancesListResponse,
    T_TypedInstanceIdentifier,
    T_WrappedInstanceRequest,
    T_WrappedInstanceResponse,
    TypedInstanceIdentifier,
    TypedNodeIdentifier,
    TypedViewReference,
)
from cognite_toolkit._cdf_tk.utils.collection import chunker_sequence

METHOD_MAP: dict[APIMethod, Endpoint] = {
    "upsert": Endpoint(method="POST", path="/models/instances", item_limit=1000),
    "retrieve": Endpoint(method="POST", path="/models/instances/byids", item_limit=1000),
    "delete": Endpoint(method="POST", path="/models/instances/delete", item_limit=1000),
    "list": Endpoint(method="POST", path="/models/instances/list", item_limit=1000),
}
QUERY_ENDPOINT = Endpoint(method="POST", path="/models/instances/query", item_limit=1000)


class InstancesAPI(CDFResourceAPI[InstanceResponse]):
    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(http_client=http_client, method_endpoint_map=METHOD_MAP)

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[InstanceResponse]:
        return PagedResponse[InstanceResponse].model_validate_json(response.body)

    def _validate_response(self, response: SuccessResponse) -> ResponseItems[TypedInstanceIdentifier]:
        return ResponseItems[TypedInstanceIdentifier].model_validate_json(response.body)

    def create(self, items: Sequence[InstanceRequest]) -> list[InstanceSlimDefinition]:
        """Create instances in CDF.

        Args:
            items: List of InstanceRequest objects to create.
        Returns:
            List of created InstanceSlimDefinition objects.
        """
        response_items: list[InstanceSlimDefinition] = []
        for response in self._chunk_requests(items, "upsert", self._serialize_items):
            response_items.extend(PagedResponse[InstanceSlimDefinition].model_validate_json(response.body).items)
        return response_items

    def retrieve(
        self, items: Sequence[TypedInstanceIdentifier], source: TypedViewReference | None = None
    ) -> list[InstanceResponse]:
        """Retrieve instances from CDF.

        Args:
            items: List of ExternalId objects to retrieve.
            source: Optional ViewReference to specify the source view for the instances.
        Returns:
            List of retrieved InstanceResponse objects.
        """
        return self._request_item_response(
            items, method="retrieve", extra_body={"sources": [{"source": source.dump()}]} if source else None
        )

    def delete(self, items: Sequence[TypedInstanceIdentifier]) -> list[TypedInstanceIdentifier]:
        """Delete instances from CDF.

        Args:
            items: List of TypedInstanceIdentifier objects to delete.
        """
        response_items: list[TypedInstanceIdentifier] = []
        for response in self._chunk_requests(items, "delete", self._serialize_items):
            response_items.extend(self._validate_response(response).items)
        return response_items

    @staticmethod
    def _create_sort_body(instance_type: Literal["node", "edge"] | None) -> list[dict]:
        """We sort by space and externalId to get a stable sort order.

        This is also more performant than sorting by using the default sort, which will sort on
        internal CDF IDs. This will be slow if you have deleted a lot of instances, as they will be counted.
        By sorting on space and externalId, we avoid this issue.
        """
        instance_type = instance_type or "node"
        return [
            {
                "property": [instance_type, "space"],
                "direction": "ascending",
            },
            {
                "property": [instance_type, "externalId"],
                "direction": "ascending",
            },
        ]

    @classmethod
    def _create_body(cls, filter: InstanceFilter | None) -> dict[str, Any]:
        return {
            **(filter.dump() if filter else {}),
            "sort": cls._create_sort_body(filter.instance_type if filter else "node"),
        }

    def paginate(
        self,
        filter: InstanceFilter | None = None,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[InstanceResponse]:
        """Iterate over all instances in CDF.

        Args:
           filter: InstanceFilter to filter instances.
            limit: Maximum number of items to return.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of InstanceResponse objects.
        """
        return self._paginate(
            cursor=cursor,
            limit=limit,
            body=self._create_body(filter),
        )

    def iterate(self, filter: InstanceFilter | None = None, limit: int = 100) -> Iterable[list[InstanceResponse]]:
        """Iterate over all instances in CDF.

        Args:
            filter: InstanceFilter to filter instances.
            limit: Maximum number of items to return per page.

        Returns:
            Iterable of lists of InstanceResponse objects.
        """
        return self._iterate(limit=limit, body=self._create_body(filter))

    def list(self, filter: InstanceFilter | None = None, limit: int | None = 100) -> list[InstanceResponse]:
        """List all instances in CDF.

        Returns:
            List of InstanceResponse objects.
        """
        return self._list(limit=limit, body=self._create_body(filter))

    @overload
    def query(self, query: QueryRequest, type_results: Literal[True]) -> QueryResponseTyped: ...

    @overload
    def query(self, query: QueryRequest, type_results: Literal[False]) -> QueryResponseUntyped: ...

    def query(self, query: QueryRequest, type_results: bool = True) -> QueryResponseTyped | QueryResponseUntyped:
        """Execute a query against the instances query endpoint.

        This uses the ``POST /models/instances/query`` endpoint which supports
        traversing the graph of nodes and edges using result set expressions.

        Args:
            query: The query request specifying what to retrieve.
                type_results: Whether to return typed results (QueryResponseTyped) or untyped results
                    (QueryResponseUntyped).

        Returns:
            QueryResult containing matching instances grouped by result set expression name.
        """
        request = RequestMessage(
            endpoint_url=self._http_client.config.create_api_url(QUERY_ENDPOINT.path),
            method=QUERY_ENDPOINT.method,
            body_content=query.dump(),
        )
        response = self._http_client.request_single_retries(request)
        success = response.get_success_or_raise()
        if type_results:
            return QueryResponseTyped.model_validate_json(success.body)
        else:
            return QueryResponseUntyped.model_validate_json(success.body)


class WrappedInstancesAPI(
    Generic[T_TypedInstanceIdentifier, T_WrappedInstanceResponse], CDFResourceAPI[T_WrappedInstanceResponse], ABC
):
    """API for wrapped instances in CDF. It is intended to be subclassed for specific wrapped instance types."""

    def __init__(self, http_client: HTTPClient, view_id: TypedViewReference) -> None:
        super().__init__(http_client=http_client, method_endpoint_map=METHOD_MAP)
        self._view_id = view_id

    @abstractmethod
    def _validate_response(self, response: SuccessResponse) -> ResponseItems[T_TypedInstanceIdentifier]:
        raise NotImplementedError()

    def create(self, items: Sequence[T_WrappedInstanceRequest]) -> list[InstanceSlimDefinition]:
        """Create instances in CDF.

        Args:
            items: List of InstanceRequest objects to create.
        Returns:
            List of created InstanceSlimDefinition objects.
        """
        response_items: list[InstanceSlimDefinition] = []
        for response in self._chunk_requests(items, "upsert", self._serialize_items):
            response_items.extend(PagedResponse[InstanceSlimDefinition].model_validate_json(response.body).items)
        return response_items

    def retrieve(self, items: Sequence[T_TypedInstanceIdentifier]) -> list[T_WrappedInstanceResponse]:
        """Retrieve instances from CDF.

        Args:
            items: List of ExternalId objects to retrieve.
            source: Optional ViewReference to specify the source view for the instances.
        Returns:
            List of retrieved InstanceResponse objects.
        """
        return self._request_item_response(
            items, method="retrieve", extra_body={"sources": [{"source": self._view_id.dump()}]}
        )

    def delete(self, items: Sequence[T_TypedInstanceIdentifier]) -> list[T_TypedInstanceIdentifier]:
        """Delete instances from CDF.

        Args:
            items: List of TypedInstanceIdentifier objects to delete.
        """
        response_items: list[T_TypedInstanceIdentifier] = []
        for response in self._chunk_requests(items, "delete", self._serialize_items):
            response_items.extend(self._validate_response(response).items)
        return response_items

    def _list_instances(
        self,
        instance_type: Literal["node", "edge"] = "node",
        spaces: list[str] | None = None,
        filter: dict[str, JsonValue] | None = None,
        limit: int | None = 100,
    ) -> list[T_WrappedInstanceResponse]:
        """List all wrapped instances in CDF.

        Args:
            instance_type: The type of instance to list. Defaults to "node".
            spaces: Optional list of spaces to filter by.
            limit: Maximum number of items to return. If None, all items are returned.

        Returns:
            List of wrapped instance response objects.
        """
        filter_ = InstanceFilter(
            instance_type=instance_type,
            space=spaces,
            source=TypedViewReference(
                space=self._view_id.space,
                external_id=self._view_id.external_id,
                version=self._view_id.version,
            ),
            filter=filter,
        )
        body = {
            **filter_.dump(),
            "sort": [
                {"property": [instance_type, "space"], "direction": "ascending"},
                {"property": [instance_type, "externalId"], "direction": "ascending"},
            ],
        }
        return self._list(limit=limit, body=body)


class MultiWrappedInstancesAPI(Generic[T_InstancesListRequest, T_InstancesListResponse], ABC):
    """API for objects that wraps multiple instances in CDF.

    For example, a Canvas is a node with edges to elements in the Canvas represented as nodes. This wrapper class
    allows creating, retrieving, updating, and deleting such objects in CDF.

    """

    def __init__(self, http_client: HTTPClient, query_chunk: int) -> None:
        self._http_client = http_client
        self._method_endpoint_map = METHOD_MAP
        self._query_chunk = query_chunk

    @abstractmethod
    def _retrieve_query(self, item: Sequence[TypedInstanceIdentifier]) -> dict[str, Any]:
        raise NotImplementedError()

    @abstractmethod
    def _validate_query_response(self, query_response: QueryResponseUntyped) -> list[T_InstancesListResponse]:
        raise NotImplementedError()

    def create(self, items: Sequence[T_InstancesListRequest]) -> list[InstanceSlimDefinition]:
        """Create instances in CDF.

        Args:
            items: List of InstanceRequest objects to create.
        Returns:
            List of created InstanceSlimDefinition objects.
        """
        endpoint = self._method_endpoint_map["upsert"]
        response_items: list[InstanceSlimDefinition] = []
        for item in items:
            instance_dicts = item.dump_instances()
            item_response: list[InstanceSlimDefinition] = []
            for chunk in chunker_sequence(instance_dicts, endpoint.item_limit):
                request = RequestMessage(
                    endpoint_url=self._http_client.config.create_api_url(endpoint.path),
                    method=endpoint.method,
                    body_content={"items": chunk},  # type: ignore[dict-item]
                )
                response = self._http_client.request_single_retries(request)
                success = response.get_success_or_raise()
                paged_response = PagedResponse[InstanceSlimDefinition].model_validate_json(success.body)
                item_response.extend(paged_response.items)
            response_items.append(self._merge_instance_slim_definitions(item_response))
        return response_items

    def _merge_instance_slim_definitions(self, items: list[InstanceSlimDefinition]) -> InstanceSlimDefinition:
        """Merge multiple InstanceSlimDefinition objects into one.

        Args:
            items: List of InstanceSlimDefinition objects to merge.
        Returns:
            Merged InstanceSlimDefinition object.
        """
        if not items:
            raise ValueError("No items to merge.")
        # Assuming all items have the same space and external_id
        base_item = items[0]
        merged_item = InstanceSlimDefinition(
            instance_type=base_item.instance_type,
            version=base_item.version,
            was_modified=any(item.was_modified for item in items),
            space=base_item.space,
            external_id=base_item.external_id,
            created_time=min(item.created_time for item in items),
            last_updated_time=max(item.last_updated_time for item in items),
        )
        return merged_item

    def update(self, items: Sequence[T_InstancesListRequest]) -> list[InstanceSlimDefinition]:
        """Update multi-wrapped instances in CDF.

        This method automatically removes underlying instances that are part of the old object but not the new one.

        Args:
            items: A sequence of multi-wrapped instance requests to update.

        Returns:
            List of updated InstanceSlimDefinition objects, one for each item in the input sequence.
        """
        endpoint = self._method_endpoint_map["upsert"]
        updated: list[InstanceSlimDefinition] = []
        for item in items:
            identifier = item.as_id()
            retrieved = self.retrieve([identifier])
            if not retrieved:
                raise ValueError(f"Item with identifier {identifier} not found for update.")
            to_delete = [id.dump() for id in (set(retrieved[0].as_ids()) - set(item.as_ids()))]
            to_update = item.dump_instances()
            item_response: list[InstanceSlimDefinition] = []
            for upsert_chunk, delete_chunk in zip_longest(
                chunker_sequence(to_update, endpoint.item_limit),
                chunker_sequence(to_delete, endpoint.item_limit),
                fillvalue=None,
            ):
                body_content: dict[str, JsonValue] = {}
                if upsert_chunk:
                    # MyPy fails do understand that list[dict[str, JsonValue]] is a subtype of JsonValue
                    body_content["items"] = upsert_chunk  # type: ignore[assignment]
                if delete_chunk:
                    body_content["delete"] = delete_chunk  # type: ignore[assignment]

                response = self._http_client.request_single_retries(
                    message=RequestMessage(
                        endpoint_url=self._http_client.config.create_api_url(endpoint.path),
                        method=endpoint.method,
                        body_content=body_content,
                    )
                )
                success = response.get_success_or_raise()
                paged_response = PagedResponse[InstanceSlimDefinition].model_validate_json(success.body)
                item_response.extend(paged_response.items)
            updated.append(self._merge_instance_slim_definitions(item_response))
        return updated

    def retrieve(self, items: Sequence[TypedNodeIdentifier]) -> list[T_InstancesListResponse]:
        """Retrieve instances from CDF.

        Args:
            items: List of ExternalId objects to retrieve.
        Returns:
            List of retrieved InstanceResponse objects.
        """
        retrieved: list[T_InstancesListResponse] = []
        for chunk in chunker_sequence(items, self._query_chunk):
            query_body = self._retrieve_query(chunk)
            request = RequestMessage(
                endpoint_url=self._http_client.config.create_api_url(QUERY_ENDPOINT.path),
                method=QUERY_ENDPOINT.method,
                body_content=query_body,
            )
            response = self._http_client.request_single_retries(request)
            success = response.get_success_or_raise()
            paged_response = QueryResponseUntyped.model_validate_json(success.body)
            retrieved.extend(self._validate_query_response(paged_response))
        return retrieved

    def delete(self, items: Sequence[TypedNodeIdentifier]) -> list[TypedNodeIdentifier]:
        """Delete instances from CDF.

        Args:
            items: List of TypedInstanceIdentifier objects to delete.
        """
        endpoint = self._method_endpoint_map["delete"]
        response_items: list[TypedNodeIdentifier] = []
        for chunk in chunker_sequence(items, endpoint.item_limit):
            request = RequestMessage(
                endpoint_url=self._http_client.config.create_api_url(endpoint.path),
                method=endpoint.method,
                body_content={"items": [item.dump() for item in chunk]},
            )
            response = self._http_client.request_single_retries(request)
            success = response.get_success_or_raise()
            validated_response = ResponseItems[TypedNodeIdentifier].model_validate_json(success.body)
            response_items.extend(validated_response.items)
        return response_items



"""Group request and response classes.

Based on the API specification at:
https://api-docs.cognite.com/20230101/tag/Groups/operation/createGroups
"""

from typing import Literal

from pydantic import Field

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    RequestResource,
    ResponseResource,
)
from cognite_toolkit._cdf_tk.client._types import Metadata
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import InternalId

from .capability import GroupCapability


class TokenAttributes(BaseModelObject):
    """Token attributes for group."""

    app_ids: list[str] | None = None


class GroupAttributes(BaseModelObject):
    """Attributes for a group."""

    token: TokenAttributes | None = None


class Group(BaseModelObject):
    """Base class for Group resources."""

    name: str
    capabilities: list[GroupCapability] | None = None
    metadata: Metadata | None = None
    attributes: GroupAttributes | None = None
    source_id: str | None = Field(None, coerce_numbers_to_str=True)
    members: list[str] | Literal["allUserAccounts"] | None = None


class GroupRequest(Group, RequestResource):
    """Group request resource for creating/updating groups."""

    id: int | None = Field(default=None, exclude=True)

    def as_id(self) -> InternalId:
        if self.id is None:
            raise ValueError("Cannot convert GroupRequest to InternalId when id is None")
        return InternalId(id=self.id)


class GroupResponse(Group, ResponseResource[GroupRequest]):
    """Group response resource returned from API."""

    id: int
    is_deleted: bool
    deleted_time: int | None = None

    @classmethod
    def request_cls(cls) -> type[GroupRequest]:
        return GroupRequest



from typing import Literal

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
)


class AuthenticationRequestDefinition(BaseModelObject):
    type: str


class AuthenticationResponseDefinition(BaseModelObject):
    type: str


class BasicAuthentication(BaseModelObject):
    type: Literal["basic"] = "basic"
    username: str


class BasicAuthenticationRequest(BasicAuthentication, AuthenticationRequestDefinition):
    password: str | None = None


class BasicAuthenticationResponse(BasicAuthentication, AuthenticationResponseDefinition): ...


class ClientCredentialAuthentication(BaseModelObject):
    type: Literal["clientCredentials"] = "clientCredentials"
    client_id: str
    token_url: str
    scopes: str
    default_expires_in: str | None = None


class ClientCredentialAuthenticationRequest(ClientCredentialAuthentication, AuthenticationRequestDefinition):
    client_secret: str


class ClientCredentialAuthenticationResponse(ClientCredentialAuthentication, AuthenticationResponseDefinition): ...


class ScramShaAuthentication(BaseModelObject):
    type: Literal["scramSha256", "scramSha512"]
    username: str


class ScramShaAuthenticationRequest(ScramShaAuthentication, AuthenticationRequestDefinition):
    password: str


class ScramShaAuthenticationResponse(ScramShaAuthentication, AuthenticationResponseDefinition): ...


class HTTPBasicAuthenticationRequest(AuthenticationRequestDefinition):
    type: Literal["header", "query"]
    key: str
    value: str


class HTTPBasicAuthenticationResponse(AuthenticationResponseDefinition):
    type: Literal["header", "query"]
    key: str



"""Data Models API for managing CDF data models.

Based on the API specification at:
https://api-docs.cognite.com/20230101/tag/Data-models/operation/createDataModels
"""

import builtins
from collections.abc import Iterable, Sequence
from typing import Any, Literal, overload

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, Endpoint, PagedResponse
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, RequestMessage, SuccessResponse
from cognite_toolkit._cdf_tk.client.request_classes.filters import DataModelFilter
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling import (
    DataModelReference,
    DataModelReferenceNoVersion,
    DataModelRequest,
    DataModelResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling._data_model import DataModelResponseWithViews


class DataModelsAPI(CDFResourceAPI[DataModelResponse]):
    """API for managing CDF data models.

    Data models use an apply/upsert pattern for create and update operations.
    """

    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "upsert": Endpoint(method="POST", path="/models/datamodels", item_limit=100),
                "retrieve": Endpoint(method="POST", path="/models/datamodels/byids", item_limit=100),
                "delete": Endpoint(method="POST", path="/models/datamodels/delete", item_limit=100),
                "list": Endpoint(method="GET", path="/models/datamodels", item_limit=1000),
            },
        )

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[DataModelResponse]:
        return PagedResponse[DataModelResponse].model_validate_json(response.body)

    def create(self, items: Sequence[DataModelRequest]) -> list[DataModelResponse]:
        """Apply (create or update) data models in CDF.

        Args:
            items: List of DataModelRequest objects to apply.

        Returns:
            List of applied DataModelResponse objects.
        """
        return self._request_item_response(items, "upsert")

    def update(self, items: Sequence[DataModelRequest]) -> list[DataModelResponse]:
        """Apply (create or update) data models in CDF.

        Args:
            items: List of DataModelRequest objects to apply.
        Returns:
            List of applied DataModelResponse objects.
        """
        return self._request_item_response(items, "upsert")

    @overload
    def retrieve(
        self, items: Sequence[DataModelReferenceNoVersion], inline_views: Literal[True]
    ) -> list[DataModelResponseWithViews]: ...

    @overload
    def retrieve(
        self, items: Sequence[DataModelReferenceNoVersion], inline_views: Literal[False] = False
    ) -> list[DataModelResponse]: ...

    def retrieve(
        self, items: Sequence[DataModelReferenceNoVersion], inline_views: bool = False
    ) -> list[DataModelResponse] | list[DataModelResponseWithViews]:
        """Retrieve data models from CDF.

        Args:
            items: List of DataModelReference objects to retrieve.
            inline_views: Whether to include full view definitions in the response.

        Returns:
            List of retrieved DataModelResponse objects.
        """
        if inline_views:
            response_items: list[DataModelResponseWithViews] = []
            for response in self._chunk_requests(
                items, "retrieve", self._serialize_items, params={"inlineViews": True}
            ):
                response_items.extend(
                    PagedResponse[DataModelResponseWithViews].model_validate_json(response.body).items
                )
            return response_items
        else:
            return self._request_item_response(items, method="retrieve", params={"inlineViews": False})

    def delete(self, items: Sequence[DataModelReference]) -> None:
        """Delete data models from CDF.

        Args:
            items: List of DataModelReference objects to delete.
        """
        self._request_no_response(items, "delete")

    def paginate(
        self,
        filter: DataModelFilter | None = None,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[DataModelResponse]:
        """Get a page of data models from CDF.

        Args:
            filter: DataModelFilter to filter data models.
            limit: Maximum number of data models to return.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of DataModelResponse objects.
        """
        return self._paginate(
            cursor=cursor,
            limit=limit,
            params=filter.dump() if filter else None,
        )

    def iterate(
        self,
        filter: DataModelFilter | None = None,
        limit: int | None = None,
    ) -> Iterable[list[DataModelResponse]]:
        """Iterate over all data models in CDF.

        Args:
            filter: DataModelFilter to filter data models.
            limit: Maximum total number of data models to return.

        Returns:
            Iterable of lists of DataModelResponse objects.
        """
        return self._iterate(
            limit=limit,
            params=filter.dump() if filter else None,
        )

    @overload
    def list(
        self,
        inline_views: Literal[True],
        filter: DataModelFilter | None = None,
        limit: int | None = None,
    ) -> list[DataModelResponseWithViews]: ...

    @overload
    def list(
        self,
        inline_views: Literal[False] = False,
        filter: DataModelFilter | None = None,
        limit: int | None = None,
    ) -> builtins.list[DataModelResponse]: ...

    def list(
        self,
        inline_views: bool = False,
        filter: DataModelFilter | None = None,
        limit: int | None = None,
    ) -> builtins.list[DataModelResponse] | builtins.list[DataModelResponseWithViews]:
        """List all data models in CDF.

        Args:
            filter: DataModelFilter to filter data models.
            limit: Maximum total number of data models to return.

        Returns:
            List of DataModelResponse objects.
        """
        if not inline_views:
            return self._list(limit=limit, params=filter.dump() if filter else None)
        else:
            response_items: list[DataModelResponseWithViews] = []
            cursor: str | None = None
            total = 0
            endpoint = self._method_endpoint_map["list"]
            while True:
                page_limit = endpoint.item_limit if limit is None else min(limit - total, endpoint.item_limit)
                parameters: dict[str, Any] = (self._filter_out_none_values(filter.dump()) or {}) if filter else {}
                parameters["inlineViews"] = True
                parameters["limit"] = page_limit
                if cursor is not None:
                    parameters["cursor"] = cursor
                request = RequestMessage(
                    endpoint_url=self._make_url(endpoint.path),
                    method=endpoint.method,
                    parameters=parameters,
                    disable_gzip=self._disable_gzip,
                    api_version=self._api_version,
                )
                response = self._http_client.request_single_retries(request).get_success_or_raise()
                page = PagedResponse[DataModelResponseWithViews].model_validate_json(response.body)
                response_items.extend(page.items)
                total += len(page.items)
                if page.next_cursor is None or (limit is not None and total >= limit):
                    break
                cursor = page.next_cursor
            return response_items



from collections.abc import Hashable, Iterable, Sequence, Sized
from typing import Any, final

from cognite.client.data_classes.capabilities import Capability

from cognite_toolkit._cdf_tk.client.request_classes.filters import (
    SimulatorModelRevisionFilter,
    SimulatorModelRoutineFilter,
    SimulatorModelRoutineRevisionFilter,
)
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId, InternalOrExternalId
from cognite_toolkit._cdf_tk.client.resource_classes.simulator_model import (
    SimulatorModelRequest,
    SimulatorModelResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.simulator_model_revision import (
    SimulatorModelRevisionRequest,
    SimulatorModelRevisionResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.simulator_routine import (
    SimulatorRoutineRequest,
    SimulatorRoutineResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.simulator_routine_revision import (
    SimulatorRoutineRevisionRequest,
    SimulatorRoutineRevisionResponse,
)
from cognite_toolkit._cdf_tk.cruds._base_cruds import ResourceCRUD
from cognite_toolkit._cdf_tk.exceptions import ToolkitNotSupported
from cognite_toolkit._cdf_tk.resource_classes import SimulatorModelYAML
from cognite_toolkit._cdf_tk.resource_classes.simulator_model_revision import SimulatorModelRevisionYAML
from cognite_toolkit._cdf_tk.resource_classes.simulator_routine import SimulatorRoutineYAML
from cognite_toolkit._cdf_tk.resource_classes.simulator_routine_revision import SimulatorRoutineRevisionYAML
from cognite_toolkit._cdf_tk.utils.diff_list import diff_list_force_hashable, diff_list_identifiable

from .data_organization import DataSetsCRUD
from .file import FileMetadataCRUD
from .timeseries import TimeSeriesCRUD


@final
class SimulatorModelCRUD(ResourceCRUD[ExternalId, SimulatorModelRequest, SimulatorModelResponse]):
    folder_name = "simulators"
    resource_cls = SimulatorModelResponse
    resource_write_cls = SimulatorModelRequest
    yaml_cls = SimulatorModelYAML
    kind = "SimulatorModel"
    dependencies = frozenset({DataSetsCRUD})
    _doc_url = "Simulator-Models/operation/create_simulator_model_simulators_models_post"

    @property
    def display_name(self) -> str:
        return "simulator models"

    @classmethod
    def get_id(cls, item: SimulatorModelRequest | SimulatorModelResponse | dict) -> ExternalId:
        if isinstance(item, dict):
            return ExternalId(external_id=item["externalId"])
        if not item.external_id:
            raise KeyError("SimulatorModel must have external_id")
        return ExternalId(external_id=item.external_id)

    @classmethod
    def get_internal_id(cls, item: SimulatorModelResponse | dict) -> int:
        if isinstance(item, dict):
            return item["id"]
        if not item.id:
            raise KeyError("SimulatorModel must have id")
        return item.id

    @classmethod
    def dump_id(cls, id: ExternalId) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def get_required_capability(
        cls, items: Sequence[SimulatorModelRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        # Simulator ACLs is not yet implemented in the PySDK, which means
        # that we cannot check for specific capabilities.
        return []

    def create(self, items: Sequence[SimulatorModelRequest]) -> list[SimulatorModelResponse]:
        return self.client.tool.simulators.models.create(items)

    def retrieve(self, ids: Sequence[ExternalId]) -> list[SimulatorModelResponse]:
        return self.client.tool.simulators.models.retrieve(list(ids), ignore_unknown_ids=True)

    def update(self, items: Sequence[SimulatorModelRequest]) -> list[SimulatorModelResponse]:
        return self.client.tool.simulators.models.update(items, mode="replace")

    def delete(self, ids: Sequence[ExternalId | InternalOrExternalId]) -> int:
        if not ids:
            return 0
        self.client.tool.simulators.models.delete(list(ids))
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[SimulatorModelResponse]:
        # Note: The SimulatorModelsAPI doesn't support data_set_external_id filtering directly,
        # so we iterate and filter in memory if needed.
        cursor: str | None = None
        data_set_id: int | None = None
        if data_set_external_id:
            data_set_id = self.client.lookup.data_sets.id(data_set_external_id, is_dry_run=False)
        while True:
            page = self.client.tool.simulators.models.paginate(
                limit=1000,
                cursor=cursor,
            )
            if data_set_id:
                # Filter by data_set_external_id in memory
                for item in page.items:
                    if item.data_set_id == data_set_id:
                        yield item
            else:
                yield from page.items
            if not page.next_cursor or not page.items:
                break
            cursor = page.next_cursor

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        """Returns all items that this item requires.

        For example, a SimulatorModel requires a DataSet, so this method would return the
        DataSetsCRUD and identifier of that dataset.
        """
        if "dataSetExternalId" in item:
            yield DataSetsCRUD, ExternalId(external_id=item["dataSetExternalId"])

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> SimulatorModelRequest:
        if ds_external_id := resource.pop("dataSetExternalId", None):
            resource["dataSetId"] = self.client.lookup.data_sets.id(ds_external_id, is_dry_run)
        return SimulatorModelRequest.model_validate(resource)

    def dump_resource(self, resource: SimulatorModelResponse, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_request_resource().dump()
        if data_set_id := dumped.pop("dataSetId", None):
            dumped["dataSetExternalId"] = self.client.lookup.data_sets.external_id(data_set_id)
        return dumped


@final
class SimulatorModelRevisionCRUD(
    ResourceCRUD[ExternalId, SimulatorModelRevisionRequest, SimulatorModelRevisionResponse]
):
    folder_name = "simulators"
    resource_cls = SimulatorModelRevisionResponse
    resource_write_cls = SimulatorModelRevisionRequest
    yaml_cls = SimulatorModelRevisionYAML
    kind = "SimulatorModelRevision"
    dependencies = frozenset({SimulatorModelCRUD, FileMetadataCRUD})
    _doc_url = "Simulator-Models/operation/create_simulator_model_revision_simulators_models_revisions_post"

    @property
    def display_name(self) -> str:
        return "simulator model revisions"

    @classmethod
    def get_id(cls, item: SimulatorModelRevisionRequest | SimulatorModelRevisionResponse | dict) -> ExternalId:
        if isinstance(item, dict):
            return ExternalId(external_id=item["externalId"])
        if not item.external_id:
            raise KeyError("SimulatorModelRevision must have external_id")
        return ExternalId(external_id=item.external_id)

    @classmethod
    def get_internal_id(cls, item: SimulatorModelRevisionResponse | dict) -> int:
        if isinstance(item, dict):
            return item["id"]
        if not item.id:
            raise KeyError("SimulatorModelRevision must have id")
        return item.id

    @classmethod
    def dump_id(cls, id: ExternalId) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def get_required_capability(
        cls, items: Sequence[SimulatorModelRevisionRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        return []

    def create(self, items: Sequence[SimulatorModelRevisionRequest]) -> list[SimulatorModelRevisionResponse]:
        return self.client.tool.simulators.model_revisions.create(items)

    def retrieve(self, ids: Sequence[ExternalId]) -> list[SimulatorModelRevisionResponse]:
        return self.client.tool.simulators.model_revisions.retrieve(list(ids), ignore_unknown_ids=True)

    def update(self, items: Sequence[SimulatorModelRevisionRequest]) -> Sized:
        # Simulator model revisions API does not support update
        raise ToolkitNotSupported(
            "You cannot update simulator model revisions. They are immutable."
            "You can change the external ID and create a new revision."
        )

    def delete(self, ids: Sequence[ExternalId | InternalOrExternalId]) -> int:
        # Simulator model revisions API does not support delete
        raise ToolkitNotSupported(
            "You cannot delete simulator model revisions. They are immutable."
            "You can delete the model - this will delete all revisions."
        )

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[SimulatorModelRevisionResponse]:
        model_external_ids: list[str] | None = None
        if parent_ids:
            model_external_ids = [str(pid) for pid in parent_ids]
        for items in self.client.tool.simulators.model_revisions.iterate(
            filter=SimulatorModelRevisionFilter(model_external_ids=model_external_ids)
        ):
            yield from items

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        if "modelExternalId" in item:
            yield SimulatorModelCRUD, ExternalId(external_id=item["modelExternalId"])
        if "fileExternalId" in item:
            yield FileMetadataCRUD, ExternalId(external_id=item["fileExternalId"])

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> SimulatorModelRevisionRequest:
        if file_external_id := resource.pop("fileExternalId", None):
            resource["fileId"] = self.client.lookup.files.id(file_external_id, is_dry_run)
        return SimulatorModelRevisionRequest.model_validate(resource)

    def dump_resource(
        self, resource: SimulatorModelRevisionResponse, local: dict[str, Any] | None = None
    ) -> dict[str, Any]:
        dumped = resource.as_request_resource().dump()
        if file_id := dumped.pop("fileId", None):
            dumped["fileExternalId"] = self.client.lookup.files.external_id(file_id)
        return dumped


@final
class SimulatorRoutineCRUD(ResourceCRUD[ExternalId, SimulatorRoutineRequest, SimulatorRoutineResponse]):
    folder_name = "simulators"
    resource_cls = SimulatorRoutineResponse
    resource_write_cls = SimulatorRoutineRequest
    yaml_cls = SimulatorRoutineYAML
    kind = "SimulatorRoutine"
    dependencies = frozenset({SimulatorModelCRUD})
    _doc_url = "Simulator-Routines/operation/create_simulator_routine_simulators_routines_post"

    support_update = False

    @property
    def display_name(self) -> str:
        return "simulator routines"

    @classmethod
    def get_id(cls, item: SimulatorRoutineRequest | SimulatorRoutineResponse | dict) -> ExternalId:
        if isinstance(item, dict):
            return ExternalId(external_id=item["externalId"])
        if not item.external_id:
            raise KeyError("SimulatorRoutine must have external_id")
        return ExternalId(external_id=item.external_id)

    @classmethod
    def get_internal_id(cls, item: SimulatorRoutineResponse | dict) -> int:
        if isinstance(item, dict):
            return item["id"]
        if not item.id:
            raise KeyError("SimulatorRoutine must have id")
        return item.id

    @classmethod
    def dump_id(cls, id: ExternalId) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def get_required_capability(
        cls, items: Sequence[SimulatorRoutineRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        return []

    def create(self, items: Sequence[SimulatorRoutineRequest]) -> list[SimulatorRoutineResponse]:
        return self.client.tool.simulators.routines.create(items)

    def retrieve(self, ids: Sequence[ExternalId]) -> list[SimulatorRoutineResponse]:
        # Simulator routines do not have a retrieve endpoint, we need to list and filter
        all_items: list[SimulatorRoutineResponse] = []
        id_set = {id_.external_id for id_ in ids}
        for batch in self.client.tool.simulators.routines.iterate(limit=None):
            for item in batch:
                if item.external_id in id_set:
                    all_items.append(item)
        return all_items

    def delete(self, ids: Sequence[ExternalId | InternalOrExternalId]) -> int:
        if not ids:
            return 0
        self.client.tool.simulators.routines.delete(list(ids))
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[SimulatorRoutineResponse]:
        model_external_ids: list[str] | None = None
        if parent_ids:
            model_external_ids = [str(pid) for pid in parent_ids]
        for items in self.client.tool.simulators.routines.iterate(
            filter=SimulatorModelRoutineFilter(model_external_ids=model_external_ids),
        ):
            yield from items

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        if "modelExternalId" in item:
            yield SimulatorModelCRUD, ExternalId(external_id=item["modelExternalId"])

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> SimulatorRoutineRequest:
        return SimulatorRoutineRequest.model_validate(resource)

    def dump_resource(self, resource: SimulatorRoutineResponse, local: dict[str, Any] | None = None) -> dict[str, Any]:
        return resource.as_request_resource().dump()


@final
class SimulatorRoutineRevisionCRUD(
    ResourceCRUD[ExternalId, SimulatorRoutineRevisionRequest, SimulatorRoutineRevisionResponse]
):
    folder_name = "simulators"
    resource_cls = SimulatorRoutineRevisionResponse
    resource_write_cls = SimulatorRoutineRevisionRequest
    yaml_cls = SimulatorRoutineRevisionYAML
    kind = "SimulatorRoutineRevision"
    dependencies = frozenset({SimulatorRoutineCRUD, TimeSeriesCRUD})
    _doc_url = "Simulator-Routines/operation/create_simulator_routine_revision_simulators_routines_revisions_post"

    @property
    def display_name(self) -> str:
        return "simulator routine revisions"

    @classmethod
    def get_id(cls, item: SimulatorRoutineRevisionRequest | SimulatorRoutineRevisionResponse | dict) -> ExternalId:
        if isinstance(item, dict):
            return ExternalId(external_id=item["externalId"])
        if not item.external_id:
            raise KeyError("SimulatorRoutineRevision must have external_id")
        return ExternalId(external_id=item.external_id)

    @classmethod
    def get_internal_id(cls, item: SimulatorRoutineRevisionResponse | dict) -> int:
        if isinstance(item, dict):
            return item["id"]
        if not item.id:
            raise KeyError("SimulatorRoutineRevision must have id")
        return item.id

    @classmethod
    def dump_id(cls, id: ExternalId) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def get_required_capability(
        cls, items: Sequence[SimulatorRoutineRevisionRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        return []

    def create(self, items: Sequence[SimulatorRoutineRevisionRequest]) -> list[SimulatorRoutineRevisionResponse]:
        return self.client.tool.simulators.routine_revisions.create(items)

    def retrieve(self, ids: Sequence[ExternalId]) -> list[SimulatorRoutineRevisionResponse]:
        return self.client.tool.simulators.routine_revisions.retrieve(list(ids), ignore_unknown_ids=True)

    def update(self, items: Sequence[SimulatorRoutineRevisionRequest]) -> Sized:
        # Simulator routine revisions API does not support update
        raise ToolkitNotSupported(
            "You cannot update simulator routine revisions. They are immutable."
            "You can change the external ID and create a new revision."
        )

    def delete(self, ids: Sequence[ExternalId | InternalOrExternalId]) -> int:
        # Simulator routine revisions API does not support delete
        raise ToolkitNotSupported(
            "You cannot delete simulator routine revisions. They are immutable."
            "You can delete the routine - this will delete all revisions."
        )

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[SimulatorRoutineRevisionResponse]:
        routine_external_ids: list[str] | None = None
        if parent_ids:
            routine_external_ids = [str(pid) for pid in parent_ids]
        for items in self.client.tool.simulators.routine_revisions.iterate(
            filter=SimulatorModelRoutineRevisionFilter(routine_external_ids=routine_external_ids)
        ):
            yield from items

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        if "routineExternalId" in item:
            yield SimulatorRoutineCRUD, ExternalId(external_id=item["routineExternalId"])
        config = item.get("configuration", {})
        if not isinstance(config, dict):
            return
        for key in ["logicalCheck", "steadyStateDetection"]:
            if isinstance(values := config.get(key), list):
                for value in values:
                    if isinstance(value, dict) and isinstance(external_id := value.get("timeseriesExternalId"), str):
                        yield TimeSeriesCRUD, ExternalId(external_id=external_id)
        for key in ["inputs", "outputs"]:
            if isinstance(io_list := config.get(key), list):
                for io_item in io_list:
                    if not isinstance(io_item, dict):
                        continue
                    if isinstance(external_id := io_item.get("saveTimeseriesExternalId"), str):
                        yield TimeSeriesCRUD, ExternalId(external_id=external_id)
                    if isinstance(external_id := io_item.get("sourceExternalId"), str):
                        yield TimeSeriesCRUD, ExternalId(external_id=external_id)

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path[0] == "configuration":
            return diff_list_force_hashable(local, cdf)
        elif json_path[0] == "script":
            return diff_list_identifiable(local, cdf, get_identifier=lambda x: x["order"])
        return super().diff_list(local, cdf, json_path)



from typing import Any, Literal

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    UpdatableRequestResource,
)
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId


class SourceRequestDefinition(UpdatableRequestResource):
    type: str
    external_id: str

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)

    def as_update(self, mode: Literal["patch", "replace"]) -> dict[str, Any]:
        output = super().as_update(mode)
        # Move type from update to top-level, as the API expects it there
        output["update"].pop("type")
        output["type"] = self.type
        return output


class SourceResponseDefinition(BaseModelObject):
    external_id: str
    created_time: int
    last_updated_time: int



from collections.abc import Iterable, Sequence
from typing import Any, Literal

from cognite_toolkit._cdf_tk.client.api.extraction_pipeline_config import ExtractionPipelineConfigsAPI
from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, PagedResponse, ResponseItems
from cognite_toolkit._cdf_tk.client.cdf_client.api import Endpoint
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, SuccessResponse
from cognite_toolkit._cdf_tk.client.request_classes.filters import ClassicFilter
from cognite_toolkit._cdf_tk.client.resource_classes.extraction_pipeline import (
    ExtractionPipelineRequest,
    ExtractionPipelineResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import InternalOrExternalId


class ExtractionPipelinesAPI(CDFResourceAPI[ExtractionPipelineResponse]):
    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "create": Endpoint(method="POST", path="/extpipes", item_limit=1000, concurrency_max_workers=1),
                "retrieve": Endpoint(method="POST", path="/extpipes/byids", item_limit=1000, concurrency_max_workers=1),
                "update": Endpoint(method="POST", path="/extpipes/update", item_limit=1000, concurrency_max_workers=1),
                "delete": Endpoint(method="POST", path="/extpipes/delete", item_limit=1000, concurrency_max_workers=1),
                "list": Endpoint(method="POST", path="/extpipes/list", item_limit=1000),
            },
        )
        self.configs = ExtractionPipelineConfigsAPI(http_client)

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[ExtractionPipelineResponse]:
        return PagedResponse[ExtractionPipelineResponse].model_validate_json(response.body)

    def _reference_response(self, response: SuccessResponse) -> ResponseItems[InternalOrExternalId]:
        return ResponseItems[InternalOrExternalId].model_validate_json(response.body)

    def create(self, items: Sequence[ExtractionPipelineRequest]) -> list[ExtractionPipelineResponse]:
        """Create extraction pipelines in CDF.

        Args:
            items: List of ExtractionPipelineRequest objects to create.
        Returns:
            List of created ExtractionPipelineResponse objects.
        """
        return self._request_item_response(items, "create")

    def retrieve(
        self, items: Sequence[InternalOrExternalId], ignore_unknown_ids: bool = False
    ) -> list[ExtractionPipelineResponse]:
        """Retrieve extraction pipelines from CDF.

        Args:
            items: List of InternalOrExternalId objects to retrieve.
            ignore_unknown_ids: Whether to ignore unknown IDs.
        Returns:
            List of retrieved ExtractionPipelineResponse objects.
        """
        return self._request_item_response(
            items, method="retrieve", extra_body={"ignoreUnknownIds": ignore_unknown_ids}
        )

    def update(
        self, items: Sequence[ExtractionPipelineRequest], mode: Literal["patch", "replace"] = "replace"
    ) -> list[ExtractionPipelineResponse]:
        """Update extraction pipelines in CDF.

        Args:
            items: List of ExtractionPipelineRequest objects to update.
            mode: Update mode, either "patch" or "replace".

        Returns:
            List of updated ExtractionPipelineResponse objects.
        """
        return self._update(items, mode=mode)

    def delete(self, items: Sequence[InternalOrExternalId], ignore_unknown_ids: bool = False) -> None:
        """Delete extraction pipelines from CDF.

        Args:
            items: List of InternalOrExternalId objects to delete.
            ignore_unknown_ids: Whether to ignore unknown IDs.
        """
        self._request_no_response(items, "delete", extra_body={"ignoreUnknownIds": ignore_unknown_ids})

    def paginate(
        self,
        filter: ClassicFilter | None = None,
        external_id_prefix: str | None = None,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[ExtractionPipelineResponse]:
        """Iterate over all extraction pipelines in CDF.

        Args:
            filter: Filter by data set IDs.
            external_id_prefix: Filter by external ID prefix.
            limit: Maximum number of items to return.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of ExtractionPipelineResponse objects.
        """
        filter_body: dict[str, Any] = filter.dump() if filter else {}
        if external_id_prefix is not None:
            filter_body["externalIdPrefix"] = external_id_prefix

        return self._paginate(
            cursor=cursor,
            limit=limit,
            body={"filter": filter_body} if filter_body else None,
        )

    def iterate(
        self, filter: ClassicFilter | None = None, external_id_prefix: str | None = None, limit: int | None = 100
    ) -> Iterable[list[ExtractionPipelineResponse]]:
        """Iterate over all extraction pipelines in CDF.

        Args:
            filter: Filter by data set IDs.
            external_id_prefix: Filter by external ID prefix.
            limit: Maximum number of items to return per page.

        Returns:
            Iterable of lists of ExtractionPipelineResponse objects.
        """
        filter_body: dict[str, Any] = filter.dump() if filter else {}
        if external_id_prefix is not None:
            filter_body["externalIdPrefix"] = external_id_prefix

        return self._iterate(
            limit=limit,
            body={"filter": filter_body} if filter_body else None,
        )

    def list(
        self,
        limit: int | None = 100,
    ) -> list[ExtractionPipelineResponse]:
        """List all extraction pipelines in CDF.

        Returns:
            List of ExtractionPipelineResponse objects.
        """
        return self._list(limit=limit)



from typing import Any, ClassVar, Literal

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    ResponseResource,
    UpdatableRequestResource,
)
from cognite_toolkit._cdf_tk.client._types import Metadata
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import DataSetId

from ._common import RobotType


class Robot(BaseModelObject):
    """Robot contains information for a single robot, including capabilities, type, video streaming setup, etc.
    These fields are updated every time a robot registers with Robotics Services and do not require manual changes.

    Args:
        name: Robot name.
        capabilities: List of external ids for the capabilities the robot can perform.
        robot_type: Type of robot.
        data_set_id: The id of the data set this asset belongs to.
        description: A brief description of the robot.
        metadata: Custom, application-specific metadata. String key -> String value.
        location_external_id: External id of the location.
    """

    name: str
    capabilities: list[str]
    robot_type: RobotType
    data_set_id: int
    description: str | None = None
    metadata: Metadata | None = None
    location_external_id: str | None = None

    def as_id(self) -> DataSetId:
        return DataSetId(data_set_id=self.data_set_id)


class RobotRequest(Robot, UpdatableRequestResource):
    """Request resource for creating or updating a Robot."""

    container_fields: ClassVar[frozenset[str]] = frozenset({"metadata"})
    non_nullable_fields: ClassVar[frozenset[str]] = frozenset({"location_external_id", "description"})

    def as_update(self, mode: Literal["patch", "replace"]) -> dict[str, Any]:
        update = super().as_update(mode)
        update["update"].pop("capabilities", None)
        return update


class RobotResponse(Robot, ResponseResource[RobotRequest]):
    """Response resource for a Robot."""

    created_time: int
    updated_time: int

    @classmethod
    def request_cls(cls) -> type[RobotRequest]:
        return RobotRequest



from collections.abc import Hashable, Iterable, Sequence
from typing import Any, final

from cognite.client.data_classes.capabilities import Capability

from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling import SpaceReference
from cognite_toolkit._cdf_tk.client.resource_classes.data_product import DataProductRequest, DataProductResponse
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId
from cognite_toolkit._cdf_tk.cruds._base_cruds import ResourceCRUD
from cognite_toolkit._cdf_tk.cruds._resource_cruds.auth import GroupAllScopedCRUD
from cognite_toolkit._cdf_tk.cruds._resource_cruds.datamodel import SpaceCRUD
from cognite_toolkit._cdf_tk.resource_classes import DataProductYAML


@final
class DataProductCRUD(ResourceCRUD[ExternalId, DataProductRequest, DataProductResponse]):
    folder_name = "data_products"
    resource_cls = DataProductResponse
    resource_write_cls = DataProductRequest
    kind = "DataProduct"
    yaml_cls = DataProductYAML
    dependencies = frozenset({GroupAllScopedCRUD, SpaceCRUD})
    support_drop = True
    support_update = True
    _doc_url = "Data-Products/operation/createDataProduct"

    @property
    def display_name(self) -> str:
        return "data products"

    @classmethod
    def get_id(cls, item: DataProductRequest | DataProductResponse | dict) -> ExternalId:
        if isinstance(item, dict):
            return ExternalId(external_id=item["externalId"])
        if isinstance(item, DataProductRequest):
            return item.as_id()
        return ExternalId(external_id=item.external_id)

    @classmethod
    def dump_id(cls, id: ExternalId) -> dict[str, Any]:
        return {"externalId": id.external_id}

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        if "schemaSpace" in item:
            yield SpaceCRUD, SpaceReference(space=item["schemaSpace"])

    @classmethod
    def get_required_capability(
        cls, items: Sequence[DataProductRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:

        # TODO: dataproductsAcl is not yet in the SDK — return empty to skip capability verification.
        # Cannot use UnknownACL due to bug in the SDK.
        # Once available, require: CREATE, READ, UPDATE, DELETE (all four actions).
        return []

    def dump_resource(self, resource: DataProductResponse, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump()
        local = local or {}
        # schemaSpace is read-only (immutable after creation).
        # Always exclude it — it cannot be updated, so comparing it only creates permanent false diffs.

        if not local.get("schemaSpace"):
            dumped.pop("schemaSpace", None)

        # Strip server-set defaults for fields not explicitly in the local YAML.
        defaults: list[tuple[str, Any]] = [
            ("isGoverned", False),
            ("description", None),
            ("tags", []),
        ]
        for key, default in defaults:
            if dumped.get(key) == default and key not in local:
                dumped.pop(key)
        return dumped

    def create(self, items: Sequence[DataProductRequest]) -> list[DataProductResponse]:
        return self.client.tool.data_products.create(list(items))

    def retrieve(self, ids: Sequence[ExternalId]) -> list[DataProductResponse]:
        return self.client.tool.data_products.retrieve(list(ids), ignore_unknown_ids=True)

    def update(self, items: Sequence[DataProductRequest]) -> list[DataProductResponse]:
        return self.client.tool.data_products.update(list(items))

    def delete(self, ids: Sequence[ExternalId]) -> int:
        if not ids:
            return 0
        self.client.tool.data_products.delete(list(ids))
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[DataProductResponse]:
        for items in self.client.tool.data_products.iterate(limit=None):
            yield from items



from ._base import ToolkitResourceRule
from ._dummy import DummyDataModelRule
from ._orchestrator import RulesOrchestrator
from ._workflow import WorkflowDatasetMissing

__all__ = ["DummyDataModelRule", "RulesOrchestrator", "ToolkitResourceRule", "WorkflowDatasetMissing"]



"""Containers API for managing CDF data modeling containers.

Based on the API specification at:
https://api-docs.cognite.com/20230101/tag/Containers/operation/ApplyContainers
"""

from collections.abc import Iterable, Sequence

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, Endpoint, PagedResponse
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, SuccessResponse
from cognite_toolkit._cdf_tk.client.request_classes.filters import ContainerFilter
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling import (
    ContainerReference,
    ContainerRequest,
    ContainerResponse,
)


class ContainersAPI(CDFResourceAPI[ContainerResponse]):
    """API for managing CDF data modeling containers.

    Containers use an apply/upsert pattern for create and update operations.
    """

    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "upsert": Endpoint(method="POST", path="/models/containers", item_limit=100),
                "retrieve": Endpoint(method="POST", path="/models/containers/byids", item_limit=100),
                "delete": Endpoint(method="POST", path="/models/containers/delete", item_limit=100),
                "list": Endpoint(method="GET", path="/models/containers", item_limit=1000),
            },
        )

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[ContainerResponse]:
        return PagedResponse[ContainerResponse].model_validate_json(response.body)

    def create(self, items: Sequence[ContainerRequest]) -> list[ContainerResponse]:
        """Create (create or update) containers in CDF.

        Args:
            items: List of ContainerRequest objects to apply.

        Returns:
            List of applied ContainerResponse objects.
        """
        return self._request_item_response(items, "upsert")

    def update(self, items: Sequence[ContainerRequest]) -> list[ContainerResponse]:
        """Update (create or update) containers in CDF.

        Args:
            items: List of ContainerRequest objects to apply.
        Returns:
            List of applied ContainerResponse objects.
        """
        return self._request_item_response(items, "upsert")

    def retrieve(self, items: Sequence[ContainerReference]) -> list[ContainerResponse]:
        """Retrieve containers from CDF.

        Args:
            items: List of ContainerReference objects to retrieve.

        Returns:
            List of retrieved ContainerResponse objects.
        """
        return self._request_item_response(items, method="retrieve")

    def delete(self, items: Sequence[ContainerReference]) -> None:
        """Delete containers from CDF.

        Args:
            items: List of ContainerReference objects to delete.
        """
        self._request_no_response(items, "delete")

    def paginate(
        self,
        filter: ContainerFilter | None = None,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[ContainerResponse]:
        """Get a page of containers from CDF.

        Args:
            filter: ContainerFilter to filter containers.
            limit: Maximum number of containers to return.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of ContainerResponse objects.
        """
        return self._paginate(
            cursor=cursor,
            limit=limit,
            params=filter.dump() if filter else None,
        )

    def iterate(
        self,
        filter: ContainerFilter | None = None,
        limit: int | None = None,
    ) -> Iterable[list[ContainerResponse]]:
        """Iterate over all containers in CDF.

        Args:
            filter: ContainerFilter to filter containers.
            limit: Maximum total number of containers to return.

        Returns:
            Iterable of lists of ContainerResponse objects.
        """
        return self._iterate(
            limit=limit,
            params=filter.dump() if filter else None,
        )

    def list(self, filter: ContainerFilter | None = None, limit: int | None = None) -> list[ContainerResponse]:
        """List all containers in CDF.

        Args:
            filter: ContainerFilter to filter containers.
            limit: Maximum total number of containers to return.

        Returns:
            List of ContainerResponse objects.
        """
        return self._list(limit=limit, params=filter.dump() if filter else None)



from cognite_toolkit._cdf_tk.commands.build_v2.data_classes._insights import Recommendation
from cognite_toolkit._cdf_tk.cruds._resource_cruds import WorkflowCRUD
from cognite_toolkit._cdf_tk.resource_classes import WorkflowYAML
from cognite_toolkit._cdf_tk.rules._base import ToolkitResourceRule

BASE_CODE = "TOOLKIT-WORKFLOW"


class WorkflowDatasetMissing(ToolkitResourceRule[WorkflowYAML]):
    """
    Checks that workflows have a dataset association.

    ## What it does
    Verifies that each workflow has a `data_set_external_id` specified, associating it with a dataset.

    ## Why is this bad?
    The dataset ID identifies which dataset a workflow is associated with. A user must have access to this
    dataset to perform any actions on the workflow, such as viewing, updating, or deleting it. Additionally,
    to manage any resources connected to the workflow (such as triggers, versions, or executions), the user
    must also have access to the dataset. Without a dataset association, access control cannot be properly
    enforced, which can lead to permission issues and security concerns.

    ## Example
    **Bad**: Workflow without dataset association
    ```yaml
    externalId: my_workflow
    # Missing data_set_external_id
    ```

    **Good**: Workflow with dataset association
    ```yaml
    externalId: my_workflow
    dataSetExternalId: my_dataset
    ```
    """

    code = f"{BASE_CODE}-001"
    resource_type = WorkflowCRUD.kind
    insight_type = Recommendation

    def validate(self) -> list[Recommendation]:

        recommendations: list[Recommendation] = []

        for resource in self.resources:
            if not resource.data_set_external_id:
                recommendations.append(
                    Recommendation(
                        message=f"{resource.as_id()!s} is missing dataset association.",
                        code=self.code,
                        fix="Add dataset association to the workflow.",
                    )
                )

        return recommendations



import builtins
from collections.abc import Iterable, Sequence
from typing import Literal

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, PagedResponse, ResponseItems
from cognite_toolkit._cdf_tk.client.cdf_client.api import Endpoint
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, SuccessResponse
from cognite_toolkit._cdf_tk.client.request_classes.filters import ClassicFilter
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling._references import InstanceId
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId, InternalId, InternalOrExternalId
from cognite_toolkit._cdf_tk.client.resource_classes.pending_instance_id import PendingInstanceId
from cognite_toolkit._cdf_tk.client.resource_classes.timeseries import TimeSeriesRequest, TimeSeriesResponse


class TimeSeriesAPI(CDFResourceAPI[TimeSeriesResponse]):
    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "create": Endpoint(method="POST", path="/timeseries", item_limit=1000, concurrency_max_workers=1),
                "retrieve": Endpoint(
                    method="POST", path="/timeseries/byids", item_limit=1000, concurrency_max_workers=1
                ),
                "update": Endpoint(
                    method="POST", path="/timeseries/update", item_limit=1000, concurrency_max_workers=1
                ),
                "delete": Endpoint(
                    method="POST", path="/timeseries/delete", item_limit=1000, concurrency_max_workers=1
                ),
                "list": Endpoint(method="POST", path="/timeseries/list", item_limit=1000),
            },
            api_version="alpha",
        )

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[TimeSeriesResponse]:
        return PagedResponse[TimeSeriesResponse].model_validate_json(response.body)

    def _reference_response(self, response: SuccessResponse) -> ResponseItems[InternalOrExternalId]:
        return ResponseItems[InternalOrExternalId].model_validate_json(response.body)

    def create(self, items: Sequence[TimeSeriesRequest]) -> list[TimeSeriesResponse]:
        """Create time series in CDF.

        Args:
            items: List of TimeSeriesRequest objects to create.
        Returns:
            List of created TimeSeriesResponse objects.
        """
        return self._request_item_response(items, "create")

    def retrieve(
        self, items: Sequence[InternalId | ExternalId | InstanceId], ignore_unknown_ids: bool = False
    ) -> list[TimeSeriesResponse]:
        """Retrieve time series from CDF.

        Args:
            items: List of InternalOrExternalId objects to retrieve.
            ignore_unknown_ids: Whether to ignore unknown IDs.
        Returns:
            List of retrieved TimeSeriesResponse objects.
        """
        return self._request_item_response(
            items, method="retrieve", extra_body={"ignoreUnknownIds": ignore_unknown_ids}
        )

    def update(
        self, items: Sequence[TimeSeriesRequest], mode: Literal["patch", "replace"] = "replace"
    ) -> list[TimeSeriesResponse]:
        """Update time series in CDF.

        Args:
            items: List of TimeSeriesRequest objects to update.
            mode: Update mode, either "patch" or "replace".

        Returns:
            List of updated TimeSeriesResponse objects.
        """
        return self._update(items, mode=mode)

    def delete(self, items: Sequence[InternalOrExternalId], ignore_unknown_ids: bool = False) -> None:
        """Delete time series from CDF.

        Args:
            items: List of InternalOrExternalId objects to delete.
            ignore_unknown_ids: Whether to ignore unknown IDs.
        """
        self._request_no_response(items, "delete", extra_body={"ignoreUnknownIds": ignore_unknown_ids})

    def paginate(
        self,
        filter: ClassicFilter | None = None,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[TimeSeriesResponse]:
        """Iterate over all time series in CDF.

        Args:
            filter: Filter by data set IDs and/or asset subtree IDs.
            limit: Maximum number of items to return.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of TimeSeriesResponse objects.
        """
        return self._paginate(
            cursor=cursor,
            limit=limit,
            body={"filter": filter.dump() if filter else None},
        )

    def iterate(
        self,
        filter: ClassicFilter | None = None,
        limit: int | None = 100,
    ) -> Iterable[list[TimeSeriesResponse]]:
        """Iterate over all time series in CDF.

        Args:
            filter: Filter by data set IDs and/or asset subtree IDs.
            limit: Maximum number of items to return per page.

        Returns:
            Iterable of lists of TimeSeriesResponse objects.
        """
        return self._iterate(
            limit=limit,
            body={"filter": filter.dump() if filter else None},
        )

    def list(
        self,
        limit: int | None = 100,
    ) -> list[TimeSeriesResponse]:
        """List all time series in CDF.

        Returns:
            List of TimeSeriesResponse objects.
        """
        return self._list(limit=limit)

    def set_pending_ids(self, items: Sequence[PendingInstanceId]) -> builtins.list[TimeSeriesResponse]:
        """Set pending instance IDs for one or more time series.

        This links asset-centric time series to DM nodes that will be created
        by the syncer service.

        Args:
            items: Sequence of PendingInstanceId objects containing the pending
                instance IDs and the time series id or external_id to link them to.

        Returns:
            List of updated TimeSeriesResponse objects.
        """
        return self._request_item_response(items, method="retrieve", endpoint="/timeseries/set-pending-instance-ids")

    def unlink_instance_ids(self, items: Sequence[InternalOrExternalId]) -> builtins.list[TimeSeriesResponse]:
        """Unlink instance IDs from time series.

        This allows a CogniteTimeSeries node in Data Modeling to be deleted
        without deleting the underlying time series data.

        Args:
            items: Sequence of InternalOrExternalId identifying the time series to unlink.

        Returns:
            List of updated TimeSeriesResponse objects.
        """
        return self._request_item_response(items, method="retrieve", endpoint="/timeseries/unlink-instance-ids")



import re
from itertools import groupby
from pathlib import Path
from typing import Any, Literal

from rich import print
from rich.panel import Panel

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.commands._base import ToolkitCommand
from cognite_toolkit._cdf_tk.commands.build_cmd import BuildCommand as OldBuildCommand
from cognite_toolkit._cdf_tk.commands.build_v2._modules_parser import ModulesParser
from cognite_toolkit._cdf_tk.commands.build_v2.build_parameters import BuildParameters
from cognite_toolkit._cdf_tk.commands.build_v2.data_classes._module import ModuleSource
from cognite_toolkit._cdf_tk.data_classes import BuildConfigYAML, BuildVariables, BuiltModuleList
from cognite_toolkit._cdf_tk.data_classes._issues import Issue, IssueList
from cognite_toolkit._cdf_tk.exceptions import ToolkitError
from cognite_toolkit._cdf_tk.tk_warnings import ToolkitWarning, WarningList
from cognite_toolkit._cdf_tk.utils.file import safe_rmtree
from cognite_toolkit._version import __version__


class BuildCommand(ToolkitCommand):
    def __init__(
        self,
        print_warning: bool = True,
        skip_tracking: bool = False,
        silent: bool = False,
        client: ToolkitClient | None = None,
    ) -> None:
        super().__init__(print_warning, skip_tracking, silent, client)
        self.issues = IssueList()

    def execute(
        self,
        verbose: bool,
        base_dir: Path,
        build_dir: Path,
        selected: list[str | Path] | None,
        build_env: str | None,
        no_clean: bool,
        client: ToolkitClient | None = None,
        on_error: Literal["continue", "raise"] = "continue",
    ) -> BuiltModuleList:
        """
        Build the resources into deployable artifacts in the build directory.
        """

        self.verbose = verbose
        self.on_error = on_error

        build_parameters = BuildParameters.load(
            organization_dir=base_dir,
            build_dir=build_dir,
            build_env_name=build_env,
            client=client,
            user_selected=selected,
        )

        # Print the build input.
        if self.verbose:
            self._print_build_input(build_parameters)

        # Tracking the project and cluster for the build.
        if build_parameters.client:
            self._additional_tracking_info.project = build_parameters.client.config.project
            self._additional_tracking_info.cluster = build_parameters.client.config.cdf_cluster

        # Load modules
        modules_parser = ModulesParser(organization_dir=base_dir, selected=selected)
        module_paths = modules_parser.parse()
        module_loading_issues = modules_parser.issues
        if module_loading_issues:
            self.issues.extend(module_loading_issues)
            self._print_or_log_issues_by_category(self.issues)
            raise ToolkitError("Module loading issues encountered. Cannot continue. See above for details.")

        # Load modules
        if module_paths:
            pass

        # modules = [Module.load(path) for path in module_paths]
        # for module in modules:
        #    continue

        # Logistics: clean and create build directory
        if prepare_issues := self._prepare_target_directory(build_dir, not no_clean):
            self.issues.extend(prepare_issues)

        # Compile the configuration and variables,
        # check syntax on module and resource level
        # for any "compilation errors and warnings"
        built_modules, build_integrity_issues = self._build_configuration(build_parameters)
        if build_integrity_issues:
            self.issues.extend(build_integrity_issues)

        # This is where we would add any recommendations for the user to improve the build.
        if build_quality_issues := self._verify_build_quality(built_modules):
            self.issues.extend(build_quality_issues)

        # Finally, print warnings grouped by category/code and location.
        self._print_or_log_issues_by_category(self.issues)

        return built_modules

    def _print_build_input(self, build_input: BuildParameters) -> None:
        print(
            Panel(
                f"Building {build_input.organization_dir!s}:\n  - Toolkit Version '{__version__!s}'\n"
                f"  - Environment name {build_input.build_env_name!r}, validation-type {build_input.config.environment.validation_type!r}.\n"
                f"  - Config '{build_input.config.filepath!s}'",
                expand=False,
            )
        )

    def _prepare_target_directory(self, build_dir: Path, clean: bool = False) -> IssueList:
        """
        Directory logistics
        """
        issues = IssueList()
        if build_dir.exists() and any(build_dir.iterdir()):
            if not clean:
                raise ToolkitError("Build directory is not empty. Run without --no-clean to remove existing files.")

            if self.verbose:
                issues.append(Issue(code="BUILD_001"))
            safe_rmtree(build_dir)
        build_dir.mkdir(parents=True, exist_ok=True)
        return issues

    def _build_configuration(self, build_input: BuildParameters) -> tuple[BuiltModuleList, IssueList]:
        issues = IssueList()
        # Use build_input.modules directly (it is already filtered by selection)

        old_build_command = OldBuildCommand(print_warning=False, skip_tracking=False)
        built_modules = old_build_command.build_config(
            build_dir=build_input.build_dir,
            organization_dir=build_input.organization_dir,
            config=build_input.config,
            packages={},
            clean=False,
            verbose=self.verbose,
            client=build_input.client,
            progress_bar=False,
            on_error=self.on_error,
        )
        # Copy tracking info from old command to self
        self._additional_tracking_info.package_ids.update(old_build_command._additional_tracking_info.package_ids)
        self._additional_tracking_info.module_ids.update(old_build_command._additional_tracking_info.module_ids)

        # Collect warnings from the old build command and convert to issues
        # Always convert warnings to issues, even if the list appears empty
        # (WarningList might have custom __bool__ behavior)
        if old_build_command.warning_list:
            converted_issues = IssueList.from_warning_list(old_build_command.warning_list)
            issues.extend(converted_issues)
        return built_modules, issues

    def _verify_build_quality(self, built_modules: BuiltModuleList) -> IssueList:
        issues = IssueList()
        return issues

    def _write(self, build_input: BuildParameters) -> None:
        # Write the build to the build directory.
        # Track lines of code built.
        raise NotImplementedError()

    def _track(self, build_input: BuildParameters) -> None:
        raise NotImplementedError()

    def _print_or_log_issues_by_category(self, issues: IssueList) -> None:
        issues_sorted = sorted(issues, key=self._issue_sort_key)
        for code, grouped_issues in groupby(issues_sorted, key=lambda issue: issue.code or ""):
            print(f"[bold]{code}[/]")
            for issue in grouped_issues:
                message = issue.message or ""
                print(f"  - {message}")

    def _issue_sort_key(self, issue: Issue) -> tuple[str, str]:
        code = issue.code or ""
        if not issue.message:
            return code, ""
        match = re.search(r"'([^']+)'", issue.message)
        path = match.group(1) if match else issue.message
        return code, path

    # Delegate to old BuildCommand for backward compatibility with tests
    def build_modules(
        self,
        modules: list[ModuleSource],
        build_dir: Path,
        variables: BuildVariables,
        verbose: bool = False,
        progress_bar: bool = False,
        on_error: Literal["continue", "raise"] = "continue",
    ) -> BuiltModuleList:
        """Delegate to old BuildCommand for backward compatibility."""
        old_cmd = OldBuildCommand()

        built_modules = old_cmd.build_modules(modules, build_dir, variables, verbose, progress_bar, on_error)  # type: ignore[arg-type]
        self._additional_tracking_info.package_ids.update(old_cmd._additional_tracking_info.package_ids)
        self._additional_tracking_info.module_ids.update(old_cmd._additional_tracking_info.module_ids)
        self.issues.extend(IssueList.from_warning_list(old_cmd.warning_list or WarningList[ToolkitWarning]()))
        return built_modules

    def build_config(
        self,
        build_dir: Path,
        organization_dir: Path,
        config: BuildConfigYAML,
        packages: dict[str, list[str]],
        clean: bool = False,
        verbose: bool = False,
        client: ToolkitClient | None = None,
        progress_bar: bool = False,
        on_error: Literal["continue", "raise"] = "continue",
    ) -> BuiltModuleList:
        """Delegate to old BuildCommand for backward compatibility."""
        old_cmd = OldBuildCommand()
        return old_cmd.build_config(
            build_dir, organization_dir, config, packages, clean, verbose, client, progress_bar, on_error
        )

    def _replace_variables(
        self,
        resource_files: list[Path],
        variables: BuildVariables,
        resource_name: str,
        module_dir: Path,
        verbose: bool = False,
    ) -> list[Any]:
        """Delegate to old BuildCommand for backward compatibility."""
        old_cmd = OldBuildCommand()
        return old_cmd._replace_variables(resource_files, variables, resource_name, module_dir, verbose)



from datetime import datetime

from pydantic import Field

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    RequestResource,
    ResponseResource,
)

from .data_modeling import DataModelReference, ViewReference


class GraphQLDataModel(BaseModelObject):
    """Base class for GraphQL data model with common fields."""

    space: str
    external_id: str
    version: str
    name: str | None = None
    description: str | None = None
    graph_ql_dml: str | None = None

    def as_id(self) -> DataModelReference:
        return DataModelReference(space=self.space, external_id=self.external_id, version=self.version)


class GraphQLDataModelRequest(GraphQLDataModel, RequestResource):
    """Request resource for creating/updating GraphQL data models."""

    previous_version: str | None = None
    preserve_dml: bool | None = None
    # Used in the loading process, but not part of the API payload.
    graphql_file: str | None = Field(None, exclude=True)


class GraphQLDataModelResponse(GraphQLDataModel, ResponseResource[GraphQLDataModelRequest]):
    """Response resource for GraphQL data models."""

    created_time: datetime
    last_updated_time: datetime
    views: list[ViewReference] | None = None

    @classmethod
    def request_cls(cls) -> type[GraphQLDataModelRequest]:
        return GraphQLDataModelRequest



import sys
from typing import Annotated, Any, Literal, TypeAlias

from pydantic import BeforeValidator, Field, PlainSerializer

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    RequestResource,
    ResponseResource,
)
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import RawDatabaseId, RawTableId
from cognite_toolkit._cdf_tk.utils.file import yaml_safe_dump

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class RAWDatabaseRequest(RequestResource):
    name: str = Field(alias="dbName")

    def as_id(self) -> RawDatabaseId:
        return RawDatabaseId(name=self.name)

    # Override dump to always use by_alias=False since the API expects name='...'
    def dump(self, camel_case: bool = True, exclude_extra: bool = False) -> dict[str, Any]:
        """Dump the resource to a dictionary.

        Args:
            camel_case (bool): Whether to use camelCase for the keys. Default is True.
            exclude_extra (bool): Whether to exclude extra fields not defined in the model. Default is False.

        """
        if exclude_extra:
            return self.model_dump(
                mode="json",
                by_alias=False,
                exclude_unset=True,
                exclude=set(self.__pydantic_extra__) if self.__pydantic_extra__ else None,
            )
        return self.model_dump(mode="json", by_alias=False, exclude_unset=True)


class RAWDatabaseResponse(ResponseResource[RAWDatabaseRequest]):
    name: str
    created_time: int

    @classmethod
    def request_cls(cls) -> type[RAWDatabaseRequest]:
        return RAWDatabaseRequest

    def as_id(self) -> RawDatabaseId:
        return RawDatabaseId(name=self.name)


class RAWTableRequest(RequestResource):
    # This is a query parameter, so we exclude it from serialization.
    db_name: str = Field(exclude=True)
    name: str = Field(alias="tableName")

    def as_id(self) -> RawTableId:
        return RawTableId(db_name=self.db_name, name=self.name)

    def dump_yaml(
        self, camel_case: bool = True, exclude_extra: bool = False, context: Literal["api", "toolkit"] = "api"
    ) -> str:
        return yaml_safe_dump(self.dump(camel_case=camel_case, exclude_extra=exclude_extra, context=context))

    # Override dump to always use by_alias=False since the API expects name='...'
    def dump(
        self, camel_case: bool = True, exclude_extra: bool = False, context: Literal["api", "toolkit"] = "api"
    ) -> dict[str, Any]:
        """Dump the resource to a dictionary.

        Args:
            camel_case (bool): Whether to use camelCase for the keys. Default is True.
            exclude_extra (bool): Whether to exclude extra fields not defined in the model. Default is False.

        """
        if exclude_extra:
            dumped = self.model_dump(
                mode="json",
                by_alias=False if context == "api" else camel_case,
                exclude_unset=True,
                exclude=set(self.__pydantic_extra__) if self.__pydantic_extra__ else None,
            )
        dumped = self.model_dump(mode="json", by_alias=False if context == "api" else camel_case, exclude_unset=True)
        if context == "api":
            return dumped
        # For toolkit context, we include dbName in the dump for consistency with other resources.
        return {**dumped, "dbName" if camel_case else "db_name": self.db_name}


class RAWTableResponse(ResponseResource[RAWTableRequest]):
    # This is a query parameter, so we exclude it from serialization.
    # Default to empty string to allow parsing from API responses (which don't include db_name).
    db_name: str = Field(default="", exclude=True)
    name: str
    created_time: int

    @classmethod
    def request_cls(cls) -> type[RAWTableRequest]:
        return RAWTableRequest

    def as_request_resource(self) -> RAWTableRequest:
        dumped = {**self.dump(), "dbName": self.db_name}
        return RAWTableRequest.model_validate(dumped, extra="ignore")

    @classmethod
    def _load(cls, resource: dict[str, Any]) -> Self:
        """Load method to match CogniteResource signature."""
        return cls.model_validate(resource, by_name=True)

    def as_id(self) -> RawTableId:
        return RawTableId(db_name=self.db_name, name=self.name)


def _unzip_pairs(value: Any) -> Any:
    """Convert parallel arrays [[a1, a2], [b1, b2]] into tuple of pairs ((a1, b1), (a2, b2)).

    Only transforms list inputs (from JSON); tuples are treated as already unpacked.
    """
    if isinstance(value, list) and len(value) == 2 and isinstance(value[0], list):
        return tuple(tuple(pair) for pair in zip(value[0], value[1]))
    return value


def _zip_pairs(value: Any) -> list[list[Any]]:
    """Convert tuple of pairs ((a1, b1), (a2, b2)) back into parallel arrays [[a1, a2], [b1, b2]]."""
    if not value:
        return [[], []]
    a, b = zip(*value)
    return [list(a), list(b)]


def _unzip_to_dict(value: Any) -> Any:
    """Convert parallel arrays [[k1, k2], [v1, v2]] into dict {k1: v1, k2: v2}.

    Only transforms list inputs (from JSON); dicts are returned as-is.
    """
    if isinstance(value, list) and len(value) == 2 and isinstance(value[0], list):
        return dict(zip(value[0], value[1]))
    return value


def _zip_from_dict(value: dict[str, int]) -> list[list[Any]]:
    """Convert dict {k1: v1, k2: v2} back into parallel arrays [[k1, k2], [v1, v2]]."""
    return [list(value.keys()), list(value.values())]


ZippedIntPairs = Annotated[
    tuple[tuple[int, int], ...],
    BeforeValidator(_unzip_pairs),
    PlainSerializer(_zip_pairs),
]
ZippedFloatIntPairs = Annotated[
    tuple[tuple[float, int], ...],
    BeforeValidator(_unzip_pairs),
    PlainSerializer(_zip_pairs),
]
ZippedStringIntDict = Annotated[
    dict[str, int],
    BeforeValidator(_unzip_to_dict),
    PlainSerializer(_zip_from_dict),
]


class StringProfile(BaseModelObject):
    length_range: tuple[int, int]
    distinct_count: int
    length_histogram: ZippedIntPairs
    value_counts: ZippedStringIntDict
    count: int


class NumberProfile(BaseModelObject):
    value_range: tuple[float, float]
    distinct_count: int
    value_counts: dict[str, int]
    histogram: ZippedFloatIntPairs
    count: int
    mean: float
    std: float
    median: float


class BooleanProfile(BaseModelObject):
    count: int
    true_count: int


class VectorProfile(BaseModelObject):
    count: int
    length_histogram: ZippedIntPairs
    length_range: tuple[int, int]


class ObjectProfile(BaseModelObject):
    count: int
    key_count_histogram: ZippedIntPairs
    key_count_range: tuple[int, int]


class RawProfileColumn(BaseModelObject):
    count: int
    null_count: int


class StringProfileColumn(RawProfileColumn):
    string: StringProfile


class NumberProfileColumn(RawProfileColumn):
    number: NumberProfile


class BooleanProfileColumn(RawProfileColumn):
    boolean: BooleanProfile


class VectorProfileColumn(RawProfileColumn):
    vector: VectorProfile


class ObjectProfileColumn(RawProfileColumn):
    object: ObjectProfile


class UnknownTypeProfileColumn(RawProfileColumn):
    """Fallback for unknown column types."""

    ...


KNOWN_COLUMN_TYPES: dict[str, type[RawProfileColumn]] = {
    "string": StringProfileColumn,
    "number": NumberProfileColumn,
    "boolean": BooleanProfileColumn,
    "vector": VectorProfileColumn,
    "object": ObjectProfileColumn,
}


def _handle_unknown_column(value: Any) -> Any:
    if isinstance(value, dict):
        for key, cls in KNOWN_COLUMN_TYPES.items():
            if key in value:
                return cls.model_validate(value)
        return UnknownTypeProfileColumn.model_validate(value)
    return value


ProfileColumnType = Annotated[
    StringProfileColumn
    | NumberProfileColumn
    | BooleanProfileColumn
    | VectorProfileColumn
    | ObjectProfileColumn
    | UnknownTypeProfileColumn,
    BeforeValidator(_handle_unknown_column),
]
RawProfileColumns: TypeAlias = dict[str, ProfileColumnType]


class RawProfileResponse(BaseModelObject):
    """The results of a raw profile operation used with the /profiler/raw endpoint.

    This endpoint is undocumented, and thus it is hard to know exactly what each field means,
    so the following is based on the experimentation with the response of the endpoint.

    Args:
        row_count: The number of rows in the profile.
        columns: The columns in the profile, each column is a RawProfileColumn.
        is_complete: Whether the profiling operation is complete or not. If not complete, the results may be partial.

    """

    row_count: int
    columns: RawProfileColumns
    is_complete: bool

    @property
    def column_count(self) -> int:
        return len(self.columns)



from collections.abc import Hashable, Iterable, Sequence, Sized
from typing import Any, final

from cognite.client.data_classes import capabilities
from cognite.client.data_classes.capabilities import Capability
from cognite.client.data_classes.data_modeling import ViewId

from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling import SpaceReference, ViewReference
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId
from cognite_toolkit._cdf_tk.client.resource_classes.instance_api import TypedNodeIdentifier
from cognite_toolkit._cdf_tk.client.resource_classes.resource_view_mapping import (
    RESOURCE_MAPPING_VIEW_ID,
    ResourceViewMappingRequest,
    ResourceViewMappingResponse,
)
from cognite_toolkit._cdf_tk.constants import COGNITE_MIGRATION_SPACE
from cognite_toolkit._cdf_tk.cruds._base_cruds import ResourceCRUD
from cognite_toolkit._cdf_tk.resource_classes import ResourceViewMappingYAML
from cognite_toolkit._cdf_tk.utils import in_dict, sanitize_filename

from .datamodel import SpaceCRUD, ViewCRUD


@final
class ResourceViewMappingCRUD(ResourceCRUD[ExternalId, ResourceViewMappingRequest, ResourceViewMappingResponse]):
    folder_name = "migration"
    resource_cls = ResourceViewMappingResponse
    resource_write_cls = ResourceViewMappingRequest
    kind = "ResourceViewMapping"
    dependencies = frozenset({SpaceCRUD, ViewCRUD})
    _doc_url = "Instances/operation/applyNodeAndEdges"
    yaml_cls = ResourceViewMappingYAML

    @property
    def display_name(self) -> str:
        return "resource view mapping"

    @classmethod
    def get_id(cls, item: ResourceViewMappingRequest | ResourceViewMappingResponse | dict) -> ExternalId:
        if isinstance(item, dict):
            return ExternalId(external_id=item["externalId"])
        return ExternalId(external_id=item.external_id)

    @classmethod
    def dump_id(cls, id: ExternalId) -> dict[str, Any]:
        return {"externalId": id.external_id}

    @classmethod
    def as_str(cls, id: ExternalId) -> str:
        return sanitize_filename(id.external_id)

    @classmethod
    def get_required_capability(
        cls, items: Sequence[ResourceViewMappingRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [capabilities.DataModelInstancesAcl.Action.Read]
            if read_only
            else [capabilities.DataModelInstancesAcl.Action.Read, capabilities.DataModelInstancesAcl.Action.Write]
        )

        return capabilities.DataModelInstancesAcl(
            actions=actions, scope=capabilities.DataModelInstancesAcl.Scope.SpaceID([COGNITE_MIGRATION_SPACE])
        )

    def prerequisite_warning(self) -> str | None:
        view_id = RESOURCE_MAPPING_VIEW_ID
        views = self.client.data_modeling.views.retrieve(
            ViewId(space=view_id.space, external_id=view_id.external_id, version=view_id.version)
        )
        if len(views) > 0:
            return None
        return f"{self.display_name} requires the {view_id!s} to be deployed. run `cdf migrate prepare` to deploy it."

    def create(self, items: Sequence[ResourceViewMappingRequest]) -> Sized:
        return self.client.migration.resource_view_mapping.create(items)

    def update(self, items: Sequence[ResourceViewMappingRequest]) -> Sized:
        return self.client.migration.resource_view_mapping.create(items)

    def retrieve(self, ids: Sequence[ExternalId]) -> list[ResourceViewMappingResponse]:
        node_ids = TypedNodeIdentifier.from_external_ids(ids, space=COGNITE_MIGRATION_SPACE)
        return self.client.migration.resource_view_mapping.retrieve(node_ids)

    def delete(self, ids: Sequence[ExternalId]) -> int:
        node_ids = TypedNodeIdentifier.from_external_ids(ids, space=COGNITE_MIGRATION_SPACE)
        result = self.client.migration.resource_view_mapping.delete(node_ids)
        return len(result)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[ResourceViewMappingResponse]:
        if space == COGNITE_MIGRATION_SPACE:
            return self.client.migration.resource_view_mapping.list(limit=-1)
        else:
            return []

    @classmethod
    def get_dependent_items(cls, item: dict) -> "Iterable[tuple[type[ResourceCRUD], Hashable]]":
        yield SpaceCRUD, SpaceReference(space=COGNITE_MIGRATION_SPACE)
        view_id = RESOURCE_MAPPING_VIEW_ID
        yield (
            ViewCRUD,
            ViewReference(space=view_id.space, external_id=view_id.external_id, version=view_id.version),
        )

        if "viewId" in item:
            view_id_dict = item["viewId"]
            if isinstance(view_id_dict, dict) and in_dict(("space", "externalId", "version"), view_id_dict):
                yield (
                    ViewCRUD,
                    ViewReference(
                        space=view_id_dict["space"],
                        external_id=view_id_dict["externalId"],
                        version=view_id_dict["version"],
                    ),
                )

    def dump_resource(
        self, resource: ResourceViewMappingResponse, local: dict[str, Any] | None = None
    ) -> dict[str, Any]:
        dumped = resource.as_request_resource().dump(context="toolkit")
        local = local or {}
        # Instance type and space is hardcoded for ResourceViewMapping,
        for key in ["existingVersion", "instanceType", "space"]:
            if key not in local:
                dumped.pop(key, None)
        if "viewId" in dumped and "viewId" in local:
            if "type" in dumped["viewId"] and "type" not in local["viewId"]:
                dumped["viewId"].pop("type", None)
        return dumped



from pydantic import Field

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    RequestResource,
    ResponseResource,
)

from .identifiers import InternalId


class TransformationNotification(BaseModelObject):
    destination: str


class TransformationNotificationRequest(TransformationNotification, RequestResource):
    transformation_external_id: str | None = None
    transformation_id: int | None = None
    # The id is not part of the create request payload, but is needed for deletion.
    id: int | None = Field(default=None, exclude=True)

    def as_id(self) -> InternalId:
        if self.id is None:
            raise ValueError("Cannot create identifier from TransformationNotificationRequest without id set")
        return InternalId(id=self.id)


class TransformationNotificationResponse(
    TransformationNotification, ResponseResource[TransformationNotificationRequest]
):
    id: int
    created_time: int
    last_updated_time: int
    transformation_id: int
    transformation_external_id: str | None = None

    @classmethod
    def request_cls(cls) -> type[TransformationNotificationRequest]:
        return TransformationNotificationRequest



from pydantic import Field

from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import NameId

from .base import ToolkitResource


class SecurityCategoriesYAML(ToolkitResource):
    name: str = Field(description="The name of the data set.")

    def as_id(self) -> NameId:
        return NameId(name=self.name)



from pathlib import Path

from cognite_toolkit._cdf_tk.constants import EXCL_FILES, MODULES
from cognite_toolkit._cdf_tk.cruds import CRUDS_BY_FOLDER_NAME, CRUDS_BY_FOLDER_NAME_INCLUDE_ALPHA
from cognite_toolkit._cdf_tk.data_classes import IssueList
from cognite_toolkit._cdf_tk.data_classes._issues import ModuleLoadingIssue
from cognite_toolkit._cdf_tk.exceptions import ToolkitError
from cognite_toolkit._cdf_tk.utils import module_path_display_name


class ModulesParser:
    def __init__(self, organization_dir: Path, selected: list[str | Path] | None = None):
        self.organization_dir = organization_dir
        self.selected = selected
        self.issues = IssueList()

    def parse(self) -> list[Path]:
        modules_root = self.organization_dir / MODULES
        if not modules_root.exists():
            raise ToolkitError(f"Module root directory '{modules_root.as_posix()}' not found")

        module_paths: list[Path] = []
        for resource_file in self.organization_dir.glob("**/*.y*ml"):
            if resource_file.name in EXCL_FILES:
                continue

            # Get the module folder for the resource file.
            module_path = self._get_module_path_from_resource_file_path(resource_file)
            if not module_path:
                continue

            # If the module has already been processed, skip it.
            if module_path in module_paths:
                continue

            module_paths.append(module_path)

        deepest_module_paths = self._find_modules_with_submodules(module_paths)
        parent_module_paths = set(module_paths) - set(deepest_module_paths)
        if parent_module_paths:
            module_paths = deepest_module_paths
            self.issues.extend(
                ModuleLoadingIssue(
                    message=f"Module {module_path_display_name(self.organization_dir, parent_module_path)!r} is skipped because it has submodules",
                )
                for parent_module_path in parent_module_paths
            )

        all_module_paths = module_paths

        if self.selected:
            normalized_selected = self._normalize_selection(self.selected)
            selected_module_paths = [
                module_path
                for module_path in all_module_paths
                if self._matches_selection(module_path, modules_root, normalized_selected)
            ]
            for selected_module in self.selected:
                normalized_selected_item = self._normalize_selection([selected_module])
                has_match = any(
                    self._matches_selection(found_module, modules_root, normalized_selected_item)
                    for found_module in all_module_paths
                )
                if not has_match:
                    self.issues.append(
                        ModuleLoadingIssue(
                            message=f"Module '{selected_module}' not found",
                        )
                    )

            module_paths = selected_module_paths

        valid_module_paths: list[Path] = []
        for module_path in module_paths:
            valid_module_path, issue = self._check_resource_folder_content(module_path)
            if issue:
                self.issues.append(issue)
            if valid_module_path:
                valid_module_paths.append(valid_module_path)

        return valid_module_paths

    def _get_module_path_from_resource_file_path(self, resource_file: Path) -> Path | None:
        # recognize the module by traversing the parents of the resource file until we find a CRUD folder

        for parent in resource_file.parents:
            if parent.name in CRUDS_BY_FOLDER_NAME_INCLUDE_ALPHA:
                # special case: if the crud is FunctionCRUD, the resource file has to be a direct descendent.
                if parent.name == "functions" and resource_file.parent.name != "functions":
                    return None
                return parent.parent
            if parent.name == MODULES:
                return parent
        return None

    def _check_resource_folder_content(self, module_path: Path) -> tuple[None | Path, ModuleLoadingIssue | None]:
        resource_folder_names = {d.name for d in module_path.iterdir() if d.is_dir()}
        unrecognized_resource_folder_names = resource_folder_names - CRUDS_BY_FOLDER_NAME.keys()

        issue = (
            ModuleLoadingIssue(
                message=f"Module {module_path_display_name(self.organization_dir, module_path)!r} contains unrecognized resource folder(s): {', '.join(unrecognized_resource_folder_names)}"
            )
            if unrecognized_resource_folder_names
            else None
        )

        has_valid_resource_folders = bool(resource_folder_names & CRUDS_BY_FOLDER_NAME.keys())
        return (module_path if has_valid_resource_folders else None, issue)

    def _matches_selection(
        self, module_path: Path, modules_root: Path, normalized_selected: list[tuple[str, ...]] | None
    ) -> bool:
        if not normalized_selected:
            return True

        rel = module_path.relative_to(modules_root)
        rel_parts = tuple(p.lower() for p in rel.parts)
        if not rel_parts:
            # module_path is the modules_root itself
            return False
        name_lower = rel_parts[-1]

        for sel_parts in normalized_selected:
            if not sel_parts:
                return True

            if len(sel_parts) == 1 and name_lower == sel_parts[0]:
                return True

            if rel_parts[: len(sel_parts)] == sel_parts:
                return True

        return False

    def _normalize_selection(self, selected: list[str | Path]) -> list[tuple[str, ...]]:
        normalized: list[tuple[str, ...]] = []
        modules_lower = MODULES.lower()
        for sel in selected:
            if isinstance(sel, Path):
                sel_parts = sel.parts
            else:
                sel_parts = tuple(part for part in str(sel).replace("\\", "/").split("/") if part)

            sel_parts_lower = tuple(part.lower() for part in sel_parts)
            if sel_parts_lower and sel_parts_lower[0] == modules_lower:
                sel_parts_lower = sel_parts_lower[1:]

            normalized.append(sel_parts_lower)

        return normalized

    def _find_modules_with_submodules(self, module_paths: list[Path]) -> list[Path]:
        """Remove parent modules when they have submodules. Keep only the deepest modules."""
        return [
            module_path
            for module_path in module_paths
            if not any(
                module_path in other_module_path.parents
                for other_module_path in module_paths
                if other_module_path != module_path
            )
        ]



from pathlib import Path
from typing import Any

import typer

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.commands._import_cmd import ImportTransformationCLI
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables


class ImportApp(typer.Typer):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.callback(invoke_without_command=True)(self.main)
        self.command("transformation-cli")(self.transformation_cli)

    def main(self, ctx: typer.Context) -> None:
        """PREVIEW FEATURE Import resources into Cognite-Toolkit."""
        if ctx.invoked_subcommand is None:
            print("Use [bold yellow]cdf-tk import --help[/] for more information.")
        return None

    @staticmethod
    def transformation_cli(
        source: Path = typer.Argument(..., help="Path to the transformation CLI manifest directory or files."),
        destination: Path = typer.Argument(..., help="Path to the destination directory."),
        overwrite: bool = typer.Option(False, help="Overwrite destination if it already exists."),
        flatten: bool = typer.Option(False, help="Flatten the directory structure."),
        clean: bool = typer.Option(False, help="Remove the source directory after import."),
        verbose: bool = typer.Option(False, help="Turn on to get more verbose output when running the command"),
    ) -> None:
        """Import transformation CLI manifests into Cognite-Toolkit modules."""

        # We are lazy loading the client as we only need it if we need to look up dataset ids.
        # This is to ensure the command can be executed without a client if the user does not need to look up dataset ids.
        # (which is likely 99% of the time)
        def get_client() -> ToolkitClient:
            return EnvironmentVariables.create_from_environment().get_client()

        cmd = ImportTransformationCLI(print_warning=True, get_client=get_client)
        cmd.execute(source, destination, overwrite, flatten, clean, verbose=verbose)



from pydantic import Field

from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId

from .base import ToolkitResource


class SimulatorModelYAML(ToolkitResource):
    """Simulator model YAML resource class.

    Based on: https://api-docs.cognite.com/20230101/tag/Simulator-Models/operation/create_simulator_model_simulators_models_post
    """

    external_id: str = Field(description="External ID of the simulator model.", min_length=1, max_length=255)
    simulator_external_id: str = Field(description="External id of the simulator.", min_length=1, max_length=50)
    name: str = Field(description="The name of the simulator model.", min_length=1, max_length=50)
    description: str | None = Field(None, description="Description of the simulator model.", max_length=500)
    data_set_external_id: str = Field(description="The external ID of the dataset this simulator model belongs to.")
    type: str = Field(description="The type of the simulator model.", min_length=1, max_length=50)

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)



from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
)


class CACertificateRequest(BaseModelObject):
    type: str
    certificate: str


class AuthCertificateRequest(BaseModelObject):
    key: str
    key_password: str | None = None
    type: str
    certificate: str


class CertificateResponse(BaseModelObject):
    thumbprint: str
    expires_at: int
